# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An implementation of Information Set Monte Carlo Tree Search (IS-MCTS).

See Cowling, Powley, and Whitehouse 2011.
https://ieeexplore.ieee.org/document/6203567
"""

import copy
import enum
import numpy as np
import pyspiel

UNLIMITED_NUM_WORLD_SAMPLES = -1
UNEXPANDED_VISIT_COUNT = -1
TIE_TOLERANCE = 1e-5


class ISMCTSFinalPolicyType(enum.Enum):
  """A enumeration class for final ISMCTS policy type."""
  NORMALIZED_VISITED_COUNT = 1
  MAX_VISIT_COUNT = 2
  MAX_VALUE = 3


class ChildSelectionPolicy(enum.Enum):
  """A enumeration class for children selection in ISMCTS."""
  UCT = 1
  PUCT = 2


class ChildInfo(object):
  """Child node information for the search tree."""

  def __init__(self, visits, return_sum, prior):
    self.visits = visits
    self.return_sum = return_sum
    self.prior = prior

  def value(self):
    return self.return_sum / self.visits


class ISMCTSNode(object):
  """Node data structure for the search tree."""

  def __init__(self):
    self.child_info = {}
    self.total_visits = 0
    self.prior_map = {}


class ISMCTSBot(pyspiel.Bot):
  """Adapted from the C++ implementation."""

  def __init__(self,
               game,
               evaluator,
               uct_c,
               max_simulations,
               max_world_samples=UNLIMITED_NUM_WORLD_SAMPLES,
               random_state=None,
               final_policy_type=ISMCTSFinalPolicyType.MAX_VISIT_COUNT,
               use_observation_string=False,
               allow_inconsistent_action_sets=False,
               child_selection_policy=ChildSelectionPolicy.PUCT):

    pyspiel.Bot.__init__(self)
    self._game = game
    self._evaluator = evaluator
    self._uct_c = uct_c
    self._max_simulations = max_simulations
    self._max_world_samples = max_world_samples
    self._final_policy_type = final_policy_type
    self._use_observation_string = use_observation_string
    self._allow_inconsistent_action_sets = allow_inconsistent_action_sets
    self._nodes = {}
    self._node_pool = []
    self._root_samples = []
    self._random_state = random_state or np.random.RandomState()
    self._child_selection_policy = child_selection_policy
    self._resampler_cb = None

  def random_number(self):
    return self._random_state.uniform()

  def reset(self):
    self._nodes = {}
    self._node_pool = []
    self._root_samples = []

  def get_state_key(self, state):
    if self._use_observation_string:
      return state.current_player(), state.observation_string()
    else:
      return state.current_player(), state.information_state_string()

  def run_search(self, state):
    self.reset()
    assert state.get_game().get_type(
    ).dynamics == pyspiel.GameType.Dynamics.SEQUENTIAL
    assert state.get_game().get_type(
    ).information == pyspiel.GameType.Information.IMPERFECT_INFORMATION

    legal_actions = state.legal_actions()
    if len(legal_actions) == 1:
      return [(legal_actions[0], 1.0)]

    self._root_node = self.create_new_node(state)

    assert self._root_node

    root_infostate_key = self.get_state_key(state)

    for _ in range(self._max_simulations):
      # how to sample a pyspiel.state from another pyspiel.state?
      sampled_root_state = self.sample_root_state(state)
      assert root_infostate_key == self.get_state_key(sampled_root_state)
      assert sampled_root_state
      self.run_simulation(sampled_root_state)

    if self._allow_inconsistent_action_sets:  # when this happens?
      legal_actions = state.legal_actions()
      temp_node = self.filter_illegals(self._root_node, legal_actions)
      assert temp_node.total_visits > 0
      return self.get_final_policy(state, temp_node)
    else:
      return self.get_final_policy(state, self._root_node)

  def step(self, state):
    action_list, prob_list = zip(*self.run_search(state))
    return self._random_state.choice(action_list, p=prob_list)

  def get_policy(self, state):
    return self.run_search(state)

  def step_with_policy(self, state):
    policy = self.get_policy(state)
    action_list, prob_list = zip(*policy)
    sampled_action = self._random_state.choice(action_list, p=prob_list)
    return policy, sampled_action

  def get_final_policy(self, state, node):
    assert node
    if (
        self._final_policy_type
        == ISMCTSFinalPolicyType.NORMALIZED_VISITED_COUNT
    ):
      assert node.total_visits > 0
      total_visits = node.total_visits
      policy = [
          (action, child.visits / total_visits)
          for action, child in node.child_info.items()
      ]
    elif self._final_policy_type == ISMCTSFinalPolicyType.MAX_VISIT_COUNT:
      assert node.total_visits > 0
      max_visits = -float('inf')
      count = 0
      for action, child in node.child_info.items():
        if child.visits == max_visits:
          count += 1
        elif child.visits > max_visits:
          max_visits = child.visits
          count = 1
      policy = [(action, 1. / count if child.visits == max_visits else 0.0)
                for action, child in node.child_info.items()]
    elif self._final_policy_type == ISMCTSFinalPolicyType.MAX_VALUE:
      assert node.total_visits > 0
      max_value = -float('inf')
      count = 0
      for action, child in node.child_info.items():
        if child.value() == max_value:
          count += 1
        elif child.value() > max_value:
          max_value = child.value()
          count = 1
      policy = [(action, 1. / count if child.value() == max_value else 0.0)
                for action, child in node.child_info.items()]

    policy_size = len(policy)
    legal_actions = state.legal_actions()
    if policy_size < len(legal_actions):  # do we really need this step?
      for action in legal_actions:
        if action not in node.child_info:
          policy.append((action, 0.0))
    return policy

  def sample_root_state(self, state):
    if self._max_world_samples == UNLIMITED_NUM_WORLD_SAMPLES:
      return self.resample_from_infostate(state)
    elif len(self._root_samples) < self._max_world_samples:
      self._root_samples.append(self.resample_from_infostate(state))
      return self._root_samples[-1].clone()
    elif len(self._root_samples) == self._max_world_samples:
      idx = self._random_state.randint(len(self._root_samples))
      return self._root_samples[idx].clone()
    else:
      raise pyspiel.SpielError(
          'Case not handled (badly set max_world_samples..?)')

  def resample_from_infostate(self, state):
    if self._resampler_cb:
      return self._resampler_cb(state, state.current_player())
    else:
      return state.resample_from_infostate(
          state.current_player(), pyspiel.UniformProbabilitySampler(0., 1.))

  def create_new_node(self, state):
    infostate_key = self.get_state_key(state)
    self._node_pool.append(ISMCTSNode())
    node = self._node_pool[-1]
    self._nodes[infostate_key] = node
    node.total_visits = UNEXPANDED_VISIT_COUNT
    return node

  def set_resampler(self, cb):
    self._resampler_cb = cb

  def lookup_node(self, state):
    if self.get_state_key(state) in self._nodes:
      return self._nodes[self.get_state_key(state)]
    return None

  def lookup_or_create_node(self, state):
    node = self.lookup_node(state)
    if node:
      return node
    return self.create_new_node(state)

  def filter_illegals(self, node, legal_actions):
    new_node = copy.deepcopy(node)
    for action, child in node.child_info.items():
      if action not in legal_actions:
        new_node.total_visits -= child.visits
        del new_node.child_info[action]
    return new_node

  def repopulate_prior_map_inconsistent_action_sets(self, state, node):
    """Repopulate the prior map for a node with an inconsistent action set.

    Args:
      state: The state of the game.
      node: The node to repopulate the prior map for.

    When we allow inconsistent action sets, it is possible that we encounter a
    legal action that was not in the prior when evaluated (when the node was)
    added to the tree. This function adds any missing actions to the prior map
    and normalizes the probabilities.
    """
    new_prior_map = node.prior_map.copy()
    for action, prob in self._evaluator.prior(state):
      if action not in new_prior_map:
        new_prior_map[action] = prob
    # now, normalize
    sum_probs = sum(new_prior_map.values())
    for action, prob in new_prior_map.items():
      new_prior_map[action] = prob / sum_probs
    node.prior_map = new_prior_map

  def expand_if_necessary(self, state, node, action):
    if action not in node.child_info:
      if self._allow_inconsistent_action_sets and action not in node.prior_map:
        # This can happen if the prior map was populated from a state that had
        # a different legal action set than the current state.
        self.repopulate_prior_map_inconsistent_action_sets(state, node)
      node.child_info[action] = ChildInfo(0.0, 0.0, node.prior_map[action])

  def select_action_tree_policy(self, state, node, legal_actions):
    if self._allow_inconsistent_action_sets:
      temp_node = self.filter_illegals(node, legal_actions)
      if temp_node.total_visits == 0:
        action = legal_actions[self._random_state.randint(
            len(legal_actions))]  # prior?
        self.expand_if_necessary(state, node, action)
        return action
      else:
        return self.select_action(temp_node)
    else:
      return self.select_action(node)

  def _action_value(self, node, child):
    assert child.visits > 0
    action_value = child.value()
    if self._child_selection_policy == ChildSelectionPolicy.UCT:
      action_value += self._uct_c * np.sqrt(
          np.log(node.total_visits) / child.visits
      )
    elif self._child_selection_policy == ChildSelectionPolicy.PUCT:
      action_value += (
          self._uct_c
          * child.prior
          * np.sqrt(node.total_visits)
          / (1 + child.visits)
      )
    else:
      raise pyspiel.SpielError('Child selection policy unrecognized.')
    return action_value

  def _select_candidate_actions(self, node):
    candidates = []

    max_action_value = max(
        [self._action_value(node, child) for child in node.child_info.values()]
    )

    # Select all the actions within the tolerance of the best action.
    for action, child in node.child_info.items():
      if self._action_value(node, child) > max_action_value - TIE_TOLERANCE:
        candidates.append(action)

    return candidates

  def select_action(self, node):
    candidates = self._select_candidate_actions(node)
    assert len(candidates) >= 1
    return candidates[self._random_state.randint(len(candidates))]

  def check_expand(self, node, legal_actions):
    if not self._allow_inconsistent_action_sets and len(
        node.child_info) == len(legal_actions):
      return pyspiel.INVALID_ACTION
    legal_actions_copy = copy.deepcopy(legal_actions)
    self._random_state.shuffle(legal_actions_copy)
    for action in legal_actions_copy:
      if action not in node.child_info:
        return action
    return pyspiel.INVALID_ACTION

  def run_simulation(self, state):
    if state.is_terminal():
      return state.returns()
    elif state.is_chance_node():
      action_list, prob_list = zip(*state.chance_outcomes())
      chance_action = self._random_state.choice(action_list, p=prob_list)
      state.apply_action(chance_action)
      return self.run_simulation(state)
    legal_actions = state.legal_actions()
    cur_player = state.current_player()
    node = self.lookup_or_create_node(state)

    assert node

    if node.total_visits == UNEXPANDED_VISIT_COUNT:
      node.total_visits = 0
      for action, prob in self._evaluator.prior(state):
        node.prior_map[action] = prob
      return self._evaluator.evaluate(state)
    else:
      chosen_action = self.check_expand(
          node, legal_actions)  # add one children at a time?
      if chosen_action != pyspiel.INVALID_ACTION:
        # check if all actions have been expanded, if not, select one?
        # if yes, ucb?
        self.expand_if_necessary(state, node, chosen_action)
      else:
        chosen_action = self.select_action_tree_policy(state, node,
                                                       legal_actions)

      assert chosen_action != pyspiel.INVALID_ACTION

      node.total_visits += 1
      node.child_info[chosen_action].visits += 1
      state.apply_action(chosen_action)
      returns = self.run_simulation(state)
      node.child_info[chosen_action].return_sum += returns[cur_player]
      return returns



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the C++ matrix game utility methods exposed to Python."""

from absl.testing import absltest

import pyspiel


class TensorGamesUtilsTest(absltest.TestCase):

  def test_extensive_to_tensor_game_type(self):
    game = pyspiel.extensive_to_tensor_game(
        pyspiel.load_game(
            "turn_based_simultaneous_game(game=blotto(players=3,coins=5))"))
    game_type = game.get_type()
    self.assertEqual(game_type.dynamics, pyspiel.GameType.Dynamics.SIMULTANEOUS)
    self.assertEqual(game_type.chance_mode,
                     pyspiel.GameType.ChanceMode.DETERMINISTIC)
    self.assertEqual(game_type.information,
                     pyspiel.GameType.Information.ONE_SHOT)
    self.assertEqual(game_type.utility, pyspiel.GameType.Utility.ZERO_SUM)

  def test_extensive_to_tensor_game_payoff_tensor(self):
    turn_based_game = pyspiel.load_game_as_turn_based(
        "blotto(players=3,coins=5)")
    tensor_game1 = pyspiel.extensive_to_tensor_game(turn_based_game)
    tensor_game2 = pyspiel.load_tensor_game("blotto(players=3,coins=5)")
    self.assertEqual(tensor_game1.shape(), tensor_game2.shape())
    s0 = turn_based_game.new_initial_state()
    self.assertEqual(tensor_game1.shape()[0], s0.num_distinct_actions())
    for a0 in range(s0.num_distinct_actions()):
      s1 = s0.child(a0)
      self.assertEqual(tensor_game1.shape()[1], s1.num_distinct_actions())
      for a1 in range(s1.num_distinct_actions()):
        s2 = s1.child(a1)
        self.assertEqual(tensor_game1.shape()[2], s2.num_distinct_actions())
        for a2 in range(s2.num_distinct_actions()):
          s3 = s2.child(a2)
          self.assertTrue(s3.is_terminal())
          for player in range(3):
            self.assertEqual(
                s3.returns()[player],
                tensor_game1.player_utility(player, (a0, a1, a2)))
            self.assertEqual(
                s3.returns()[player],
                tensor_game2.player_utility(player, (a0, a1, a2)))


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example algorithm to sample some states from a game."""

import random
import pyspiel


def sample_some_states(
    game,
    max_states=100,
    make_distribution_fn=lambda states: [1 / len(states)] * len(states)):
  """Samples some states in the game.

  This can be run for large games, in contrast to `get_all_states`. It is useful
  for tests that need to check a predicate only on a subset of the game, since
  generating the whole game is infeasible.

  Currently only works for sequential games. For simultaneous games and mean
  field games it returns only the initial state.

  The algorithm maintains a list of states and repeatedly picks a random state
  from the list to expand until enough states have been sampled.

  Arguments:
    game: The game to analyze, as returned by `load_game`.
    max_states: The maximum number of states to return. Negative means no limit.
    make_distribution_fn: Function that takes a list of states and returns a
      corresponding distribution (as a list of floats). Only used for mean field
      games.

  Returns:
    A `list` of `pyspiel.State`.
  """
  if game.get_type().dynamics in [
      pyspiel.GameType.Dynamics.SIMULTANEOUS,
      pyspiel.GameType.Dynamics.MEAN_FIELD
  ]:
    return [game.new_initial_state()]
  states = []
  unexplored_actions = []
  indexes_with_unexplored_actions = set()

  def add_state(state):
    states.append(state)
    if state.is_terminal():
      unexplored_actions.append(None)
    else:
      indexes_with_unexplored_actions.add(len(states) - 1)
      unexplored_actions.append(set(state.legal_actions()))

  def expand_random_state():
    index = random.choice(list(indexes_with_unexplored_actions))
    state = states[index]
    if state.is_mean_field_node():
      child = state.clone()
      child.update_distribution(
          make_distribution_fn(child.distribution_support()))
      indexes_with_unexplored_actions.remove(index)
      return child
    else:
      actions = unexplored_actions[index]
      assert actions, f"Empty actions for state {state}"
      action = random.choice(list(actions))
      actions.remove(action)
      if not actions:
        indexes_with_unexplored_actions.remove(index)
      return state.child(action)

  add_state(game.new_initial_state())
  while (len(states) < max_states) and indexes_with_unexplored_actions:
    add_state(expand_random_state())

  if not states:
    raise ValueError("get_some_states sampled 0 states!")

  return states



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.efr."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import efr
from open_spiel.python.algorithms import expected_game_score
import pyspiel


class EFRTest(parameterized.TestCase, absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.kuhn_game = pyspiel.load_game("kuhn_poker")
    self.leduc_game = pyspiel.load_game("leduc_poker")
    self.kuhn_3p_game = pyspiel.load_game("kuhn_poker(players=3)")
    self.sheriff_game = pyspiel.load_game("sheriff")

    self.kuhn_uniform_policy = policy.TabularPolicy(self.kuhn_game)
    self.leduc_uniform_policy = policy.TabularPolicy(self.leduc_game)

  @parameterized.parameters([
      "blind action",
      "informed action",
      "blind cf",
      "informed cf",
      "bps",
      "cfps",
      "csps",
      "tips",
      "bhv",
  ])
  def test_policy_zero_is_uniform(self, deviations_name):
    # We use Leduc and not Kuhn, because Leduc has illegal actions and Kuhn does
    # not.
    cfr_solver = efr.EFRSolver(
        game=self.leduc_game, deviations_name=deviations_name
    )
    np.testing.assert_array_equal(
        self.leduc_uniform_policy.action_probability_array,
        cfr_solver.current_policy().action_probability_array,
    )
    np.testing.assert_array_equal(
        self.leduc_uniform_policy.action_probability_array,
        cfr_solver.average_policy().action_probability_array,
    )

  @parameterized.parameters(
      ["blind cf", "informed cf", "bps", "cfps", "csps", "tips", "bhv"]
  )
  def test_efr_kuhn_poker(self, deviations_name):
    efr_solver = efr.EFRSolver(
        game=self.kuhn_game, deviations_name=deviations_name
    )
    for _ in range(300):
      efr_solver.evaluate_and_update_policy()
    average_policy = efr_solver.average_policy()
    average_policy_values = expected_game_score.policy_value(
        self.kuhn_game.new_initial_state(), [average_policy] * 2
    )
    # 1/18 is the Nash value. See https://en.wikipedia.org/wiki/Kuhn_poker
    np.testing.assert_allclose(
        average_policy_values, [-1 / 18, 1 / 18], atol=1e-3
    )

  @parameterized.parameters(
      ["blind cf", "informed cf", "bps", "cfps", "csps", "tips", "bhv"]
  )
  def test_efr_kuhn_poker_3p(self, deviations_name):
    efr_solver = efr.EFRSolver(
        game=self.kuhn_3p_game, deviations_name=deviations_name
    )
    strategies = []
    corr_dist_values = []
    for _ in range(10):
      efr_solver.evaluate_and_update_policy()
      # Convert the policy to a pyspiel.TabularPolicy, needed by the CorrDist
      # functions on the C++ side.
      strategies.append(
          policy.python_policy_to_pyspiel_policy(efr_solver.current_policy())
      )
      corr_dev = pyspiel.uniform_correlation_device(strategies)
      cce_dist_info = pyspiel.cce_dist(self.kuhn_3p_game, corr_dev)
      corr_dist_values.append(cce_dist_info.dist_value)
    self.assertLess(corr_dist_values[-1], corr_dist_values[0])

  @absltest.skip("Too long for a unit test")
  @parameterized.parameters(["blind cf", "bps", "tips"])
  def test_efr_cce_dist_sheriff(self, deviations_name):
    efr_solver = efr.EFRSolver(
        game=self.sheriff_game, deviations_name=deviations_name
    )
    strategies = []
    corr_dist_values = []
    for _ in range(5):
      efr_solver.evaluate_and_update_policy()
      strategies.append(
          policy.python_policy_to_pyspiel_policy(efr_solver.current_policy())
      )
      corr_dev = pyspiel.uniform_correlation_device(strategies)
      cce_dist_info = pyspiel.cce_dist(self.sheriff_game, corr_dev)
      corr_dist_values.append(cce_dist_info.dist_value)
    self.assertLess(corr_dist_values[-1], corr_dist_values[0])


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for dynamic_routing_to_mean_field_game."""
from absl.testing import absltest

from open_spiel.python import games  # pylint:disable=unused-import
from open_spiel.python import policy
from open_spiel.python.algorithms import expected_game_score
from open_spiel.python.games import dynamic_routing_to_mean_field_game
from open_spiel.python.mfg import games as mfg_games  # pylint:disable=unused-import
from open_spiel.python.mfg.algorithms import mirror_descent
import pyspiel


class DerivedNPlayerPolicyFromMeanFieldPolicyTest(absltest.TestCase):

  def test_state_conversion_method(self):
    """Test N player game state to mean field game state conversion."""
    # Test state conversion.

  def test_uniform_mfg_policy_conversion_to_n_player_uniform_policy(self):
    """Test conversion of uniform to uniform policy."""
    mfg_game = pyspiel.load_game("python_mfg_dynamic_routing", {
        "time_step_length": 0.05,
        "max_num_time_step": 100
    })
    n_player_game = pyspiel.load_game("python_dynamic_routing", {
        "time_step_length": 0.05,
        "max_num_time_step": 100
    })
    mfg_derived_policy = (
        dynamic_routing_to_mean_field_game
        .DerivedNPlayerPolicyFromMeanFieldPolicy(
            n_player_game, policy.UniformRandomPolicy(mfg_game)))
    derived_policy_value = expected_game_score.policy_value(
        n_player_game.new_initial_state(), mfg_derived_policy)
    uniform_policy_value = expected_game_score.policy_value(
        n_player_game.new_initial_state(),
        policy.UniformRandomPolicy(n_player_game))
    self.assertSequenceAlmostEqual(derived_policy_value, uniform_policy_value)

  def test_pigou_network_game_outcome_optimal_mfg_policy_in_n_player_game(self):
    """Test MFG Nash equilibrium policy for the Pigou network."""
    # Test policy.
    # Test game outcome.

  def test_learning_and_applying_mfg_policy_in_n_player_game(self):
    """Test converting learnt MFG policy default game."""
    # learning the Braess MFG Nash equilibrium
    mfg_game = pyspiel.load_game("python_mfg_dynamic_routing")
    omd = mirror_descent.MirrorDescent(mfg_game, lr=1)
    for _ in range(10):
      omd.iteration()
    mfg_policy = omd.get_policy()
    n_player_game = pyspiel.load_game("python_dynamic_routing")
    mfg_derived_policy = (
        dynamic_routing_to_mean_field_game
        .DerivedNPlayerPolicyFromMeanFieldPolicy(n_player_game, mfg_policy))
    expected_game_score.policy_value(n_player_game.new_initial_state(),
                                     mfg_derived_policy)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Miscellaneous utils."""

from absl import logging  # pylint:disable=unused-import

import numpy as np


def uniform_dist(x):
  """Returns a uniform distribution with same shape as the given numpy array.

  Args:
    x: numpy array
  Returns:
    constant numpy array of same shape as input x, sums to 1
  """
  return np.ones_like(x) / float(x.size)


def argmax(random, z):
  """Returns argmax of flattened z with ties split randomly.

  Args:
    random: Random number generator, e.g., np.random.RandomState()
    z: np.array
  Returns:
    integer representing index of argmax
  """
  inds = np.arange(z.size)
  random.shuffle(inds)
  z_shuffled = z[inds]
  ind_max = np.argmax(z_shuffled)
  return inds[ind_max]


def pt_reduce(payoff_tensor, strats, remove_players):
  """Computes possible payoffs for remove_players with others' strats fixed.

  This is equivalent to the Jacobian of the payoff w.r.t. remove_players:
  sum_{a...z} A_k * x_1a * ... * x_nz for player k.
  Args:
    payoff_tensor: a single player k's payoff tensor, i.e.,
      a num action x ... x num action (num player) np.array
    strats: list of distributions over strategies for each player
    remove_players: players to NOT sum over in expectation
  Returns:
    payoff tensor of shape: num_action x ... x num_action,
      num_action for each player in remove_players
  """
  result = np.copy(payoff_tensor)
  result_dims = list(range(len(result.shape)))
  other_player_idxs = list(result_dims)
  for remove_player in remove_players:
    other_player_idxs.remove(remove_player)
  for other_player_idx in other_player_idxs:
    new_result_dims = list(result_dims)
    new_result_dims.remove(other_player_idx)
    result = np.einsum(result, result_dims, strats[other_player_idx],
                       [other_player_idx], new_result_dims)
    result_dims = new_result_dims
  return result


def isnan(x):
  """Checks for NaN's in nested objects."""
  if isinstance(x, float):
    return np.isnan(x)
  elif isinstance(x, int):
    return np.isnan(x)
  elif isinstance(x, np.ndarray):
    return np.any(np.isnan(x))
  elif isinstance(x, list):
    return np.any([isnan(xi) for xi in x])
  elif isinstance(x, tuple):
    return np.any([isnan(xi) for xi in x])
  elif isinstance(x, dict):
    return np.any([isnan(xi) for xi in x.values()])
  else:
    typ = repr(type(x))
    err_string = 'type(x)={:s} not recognized when checking for NaN'.format(typ)
    raise NotImplementedError(err_string)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Hedge algorithm for MFGs."""

from typing import Optional

import numpy as np

from open_spiel.python.mfg.algorithms import utils
from open_spiel.python.mfg.algorithms.regret import polynomial_weights


class Hedge(polynomial_weights.PolynomialWeightAlgorithm):
  """Hedge algorithm."""

  def __init__(
      self,
      game,
      policies,
      eta: Optional[float] = None,
      regret_steps_per_step: int = 1,
      rho_tol: float = 1e-4,
      compress_nus: bool = True,
      compress_lbd: float = 0.0,
      compress_every: int = 1,
      stop_early: bool = True,
      stop_regret_threshold: float = 1e-3,
      value_estimator=utils.sample_value,
      value_estimation_n: int = 1,
      compute_internal_regret: bool = False,
  ):
    super().__init__(
        game,
        policies,
        eta=eta,
        regret_steps_per_step=regret_steps_per_step,
        rho_tol=rho_tol,
        compress_nus=compress_nus,
        compress_every=compress_every,
        compress_lbd=compress_lbd,
        stop_early=stop_early,
        stop_regret_threshold=stop_regret_threshold,
        value_estimator=value_estimator,
        value_estimation_n=value_estimation_n,
        compute_internal_regret=compute_internal_regret,
    )

    if self._compute_internal_regret:
      self._ws = [np.ones(len(policies)) for _ in range(len(policies))]
      self._p = np.ones(len(policies)) / (1.0 * len(policies))
    else:
      self._w = np.ones(len(policies))

    if eta is None:
      assert regret_steps_per_step is not None, (
          "Both `eta` and "
          "`regret_steps_per_step` were "
          "None, whereas our algorithm "
          "requires either value to be "
          "set."
      )
      self.compute_optimal_eta()
      self._constant_eta = 1.0
    else:
      self._eta = eta
      self._constant_eta = eta

  def _update_weights(self, rewards):
    if self._compute_internal_regret:
      self._ws = [
          w * np.exp(self._eta * rewards * p) for w, p in zip(self._ws, self._p)
      ]
      self.compute_p()
    else:
      self._w = self._w * np.exp(self._eta * rewards)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3.
r"""Policy Gradient based agents implemented in TensorFlow.

This class is composed of three policy gradient (PG) algorithms:

- Q-based Policy Gradient (QPG): an "all-actions" advantage actor-critic
algorithm differing from A2C in that all action values are used to estimate the
policy gradient (as opposed to only using the action taken into account):

    baseline = \sum_a pi_a * Q_a
    loss = - \sum_a pi_a * (Q_a - baseline)

where (Q_a - baseline) is the usual advantage. QPG is also known as Mean
Actor-Critic (https://arxiv.org/abs/1709.00503).


- Regret policy gradient (RPG): a PG algorithm inspired by counterfactual regret
minimization (CFR). Unlike standard actor-critic methods (e.g. A2C), the loss is
defined purely in terms of thresholded regrets as follows:

    baseline = \sum_a pi_a * Q_a
    loss = regret = \sum_a relu(Q_a - baseline)

where gradients only flow through the action value (Q_a) part and are blocked on
the baseline part (which is trained separately by usual MSE loss).
The lack of negative sign in the front of the loss represents a switch from
gradient ascent on the score to descent on the loss.


- Regret Matching Policy Gradient (RMPG): inspired by regret-matching, the
policy gradient is by weighted by the thresholded regret:

    baseline = \sum_a pi_a * Q_a
    loss = - \sum_a pi_a * relu(Q_a - baseline)


These algorithms were published in NeurIPS 2018. Paper title: "Actor-Critic
Policy Optimization in Partially Observable Multiagent Environment", the paper
is available at: https://arxiv.org/abs/1810.09026.

- Advantage Actor Critic (A2C): The popular advantage actor critic (A2C)
algorithm. The algorithm uses the baseline (Value function) as a control variate
to reduce variance of the policy gradient. The loss is only computed for the
actions actually taken in the episode as opposed to a loss computed for all
actions in the variants above.

  advantages = returns - baseline
  loss = -log(pi_a) * advantages

The algorithm can be found in the textbook:
https://incompleteideas.net/book/RLbook2018.pdf under the chapter on
`Policy Gradients`.

See  open_spiel/python/algorithms/losses/rl_losses_test.py for an example of the
loss computation.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import collections
import os
from absl import logging
import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python import rl_agent
from open_spiel.python import simple_nets
from open_spiel.python.algorithms.losses import rl_losses

# Temporarily disable TF2 behavior until we update the code.
tf.disable_v2_behavior()

Transition = collections.namedtuple(
    "Transition", "info_state action reward discount legal_actions_mask")


class PolicyGradient(rl_agent.AbstractAgent):
  """RPG Agent implementation in TensorFlow.

  See open_spiel/python/examples/single_agent_catch.py for an usage example.
  """

  def __init__(self,
               session,
               player_id,
               info_state_size,
               num_actions,
               loss_str="a2c",
               loss_class=None,
               hidden_layers_sizes=(128,),
               batch_size=16,
               critic_learning_rate=0.01,
               pi_learning_rate=0.001,
               entropy_cost=0.01,
               num_critic_before_pi=8,
               additional_discount_factor=1.0,
               max_global_gradient_norm=None,
               optimizer_str="sgd"):
    """Initialize the PolicyGradient agent.

    Args:
      session: Tensorflow session.
      player_id: int, player identifier. Usually its position in the game.
      info_state_size: int, info_state vector size.
      num_actions: int, number of actions per info state.
      loss_str: string or None. If string, must be one of ["rpg", "qpg", "rm",
        "a2c"] and defined in `_get_loss_class`. If None, a loss class must be
        passed through `loss_class`. Defaults to "a2c".
      loss_class: Class or None. If Class, it must define the policy gradient
        loss. If None a loss class in a string format must be passed through
        `loss_str`. Defaults to None.
      hidden_layers_sizes: iterable, defines the neural network layers. Defaults
          to (128,), which produces a NN: [INPUT] -> [128] -> ReLU -> [OUTPUT].
      batch_size: int, batch size to use for Q and Pi learning. Defaults to 128.
      critic_learning_rate: float, learning rate used for Critic (Q or V).
        Defaults to 0.01.
      pi_learning_rate: float, learning rate used for Pi. Defaults to 0.001.
      entropy_cost: float, entropy cost used to multiply the entropy loss. Can
        be set to None to skip entropy computation. Defaults to 0.01.
      num_critic_before_pi: int, number of Critic (Q or V) updates before each
        Pi update. Defaults to 8 (every 8th critic learning step, Pi also
        learns).
      additional_discount_factor: float, additional discount to compute returns.
        Defaults to 1.0, in which case, no extra discount is applied.  None that
        users must provide *only one of* `loss_str` or `loss_class`.
      max_global_gradient_norm: float or None, maximum global norm of a gradient
        to which the gradient is shrunk if its value is larger. Defaults to
        None.
      optimizer_str: String defining which optimizer to use. Supported values
        are {sgd, adam}. Defaults to sgd
    """
    assert bool(loss_str) ^ bool(loss_class), "Please provide only one option."
    self._kwargs = locals()
    loss_class = loss_class if loss_class else self._get_loss_class(loss_str)
    self._loss_class = loss_class

    self.player_id = player_id
    self._session = session
    self._num_actions = num_actions
    self._layer_sizes = hidden_layers_sizes
    self._batch_size = batch_size
    self._extra_discount = additional_discount_factor
    self._num_critic_before_pi = num_critic_before_pi

    self._episode_data = []
    self._dataset = collections.defaultdict(list)
    self._prev_time_step = None
    self._prev_action = None

    # Step counters
    self._step_counter = 0
    self._episode_counter = 0
    self._num_learn_steps = 0

    # Keep track of the last training loss achieved in an update step.
    self._last_loss_value = None

    # Placeholders
    self._info_state_ph = tf.placeholder(
        shape=[None, info_state_size], dtype=tf.float32, name="info_state_ph")
    self._action_ph = tf.placeholder(
        shape=[None], dtype=tf.int32, name="action_ph")
    self._return_ph = tf.placeholder(
        shape=[None], dtype=tf.float32, name="return_ph")

    # Network
    # activate final as we plug logit and qvalue heads afterwards.
    self._net_torso = simple_nets.MLPTorso(info_state_size, self._layer_sizes)
    torso_out = self._net_torso(self._info_state_ph)
    torso_out_size = self._layer_sizes[-1]
    self._policy_logits_layer = simple_nets.Linear(
        torso_out_size,
        self._num_actions,
        activate_relu=False,
        name="policy_head")
    # Do not remove policy_logits_network. Even if it's not used directly here,
    # other code outside this file refers to it.
    self.policy_logits_network = simple_nets.Sequential(
        [self._net_torso, self._policy_logits_layer])
    self._policy_logits = self._policy_logits_layer(torso_out)
    self._policy_probs = tf.nn.softmax(self._policy_logits)

    self._savers = []

    # Add baseline (V) head for A2C (or Q-head for QPG / RPG / RMPG)
    if loss_class.__name__ == "BatchA2CLoss":
      self._baseline_layer = simple_nets.Linear(
          torso_out_size, 1, activate_relu=False, name="baseline")
      self._baseline = tf.squeeze(self._baseline_layer(torso_out), axis=1)
    else:
      self._q_values_layer = simple_nets.Linear(
          torso_out_size,
          self._num_actions,
          activate_relu=False,
          name="q_values_head")
      self._q_values = self._q_values_layer(torso_out)

    # Critic loss
    # Baseline loss in case of A2C
    if loss_class.__name__ == "BatchA2CLoss":
      self._critic_loss = tf.reduce_mean(
          tf.losses.mean_squared_error(
              labels=self._return_ph, predictions=self._baseline))
    else:
      # Q-loss otherwise.
      action_indices = tf.stack(
          [tf.range(tf.shape(self._q_values)[0]), self._action_ph], axis=-1)
      value_predictions = tf.gather_nd(self._q_values, action_indices)
      self._critic_loss = tf.reduce_mean(
          tf.losses.mean_squared_error(
              labels=self._return_ph, predictions=value_predictions))
    if optimizer_str == "adam":
      self._critic_optimizer = tf.train.AdamOptimizer(
          learning_rate=critic_learning_rate)
    elif optimizer_str == "sgd":
      self._critic_optimizer = tf.train.GradientDescentOptimizer(
          learning_rate=critic_learning_rate)
    else:
      raise ValueError("Not implemented, choose from 'adam' and 'sgd'.")

    def minimize_with_clipping(optimizer, loss):
      grads_and_vars = optimizer.compute_gradients(loss)
      if max_global_gradient_norm is not None:
        grads, variables = zip(*grads_and_vars)
        grads, _ = tf.clip_by_global_norm(grads, max_global_gradient_norm)
        grads_and_vars = list(zip(grads, variables))

      return optimizer.apply_gradients(grads_and_vars)

    self._critic_learn_step = minimize_with_clipping(self._critic_optimizer,
                                                     self._critic_loss)

    # Pi loss
    pg_class = loss_class(entropy_cost=entropy_cost)
    if loss_class.__name__ == "BatchA2CLoss":
      self._pi_loss = pg_class.loss(
          policy_logits=self._policy_logits,
          baseline=self._baseline,
          actions=self._action_ph,
          returns=self._return_ph)
    else:
      self._pi_loss = pg_class.loss(
          policy_logits=self._policy_logits, action_values=self._q_values)
    if optimizer_str == "adam":
      self._pi_optimizer = tf.train.AdamOptimizer(
          learning_rate=pi_learning_rate)
    elif optimizer_str == "sgd":
      self._pi_optimizer = tf.train.GradientDescentOptimizer(
          learning_rate=pi_learning_rate)

    self._pi_learn_step = minimize_with_clipping(self._pi_optimizer,
                                                 self._pi_loss)
    self._loss_str = loss_str
    self._initialize()

  def _get_loss_class(self, loss_str):
    if loss_str == "rpg":
      return rl_losses.BatchRPGLoss
    elif loss_str == "qpg":
      return rl_losses.BatchQPGLoss
    elif loss_str == "rm":
      return rl_losses.BatchRMLoss
    elif loss_str == "a2c":
      return rl_losses.BatchA2CLoss

  def _act(self, info_state, legal_actions):
    # Make a singleton batch for NN compatibility: [1, info_state_size]
    info_state = np.reshape(info_state, [1, -1])
    policy_probs = self._session.run(
        self._policy_probs, feed_dict={self._info_state_ph: info_state})

    # Remove illegal actions, re-normalize probs
    probs = np.zeros(self._num_actions)
    probs[legal_actions] = policy_probs[0][legal_actions]
    if sum(probs) != 0:
      probs /= sum(probs)
    else:
      probs[legal_actions] = 1 / len(legal_actions)
    action = np.random.choice(len(probs), p=probs)
    return action, probs

  def step(self, time_step, is_evaluation=False):
    """Returns the action to be taken and updates the network if needed.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: bool, whether this is a training or evaluation call.
          Defaults to False.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """
    # Act step: don't act at terminal info states or if its not our turn.
    if (not time_step.last()) and (
        time_step.is_simultaneous_move() or
        self.player_id == time_step.current_player()):
      info_state = time_step.observations["info_state"][self.player_id]
      legal_actions = time_step.observations["legal_actions"][self.player_id]
      action, probs = self._act(info_state, legal_actions)
    else:
      action = None
      probs = []

    if not is_evaluation:
      self._step_counter += 1

      # Add data points to current episode buffer.
      if self._prev_time_step:
        self._add_transition(time_step)

      # Episode done, add to dataset and maybe learn.
      if time_step.last():
        self._add_episode_data_to_dataset()
        self._episode_counter += 1

        if len(self._dataset["returns"]) >= self._batch_size:
          self._critic_update()
          self._num_learn_steps += 1
          if self._num_learn_steps % self._num_critic_before_pi == 0:
            self._pi_update()
          self._dataset = collections.defaultdict(list)

        self._prev_time_step = None
        self._prev_action = None
        return
      else:
        self._prev_time_step = time_step
        self._prev_action = action

    return rl_agent.StepOutput(action=action, probs=probs)

  def _full_checkpoint_name(self, checkpoint_dir, name):
    checkpoint_filename = "_".join(
        [self._loss_str, name, "pid" + str(self.player_id)])
    return os.path.join(checkpoint_dir, checkpoint_filename)

  def _latest_checkpoint_filename(self, name):
    checkpoint_filename = "_".join(
        [self._loss_str, name, "pid" + str(self.player_id)])
    return checkpoint_filename + "_latest"

  def save(self, checkpoint_dir):
    for name, saver in self._savers:
      path = saver.save(
          self._session,
          self._full_checkpoint_name(checkpoint_dir, name),
          latest_filename=self._latest_checkpoint_filename(name))
      logging.info("saved to path: %s", path)

  def has_checkpoint(self, checkpoint_dir):
    for name, _ in self._savers:
      if tf.train.latest_checkpoint(
          self._full_checkpoint_name(checkpoint_dir, name),
          os.path.join(checkpoint_dir,
                       self._latest_checkpoint_filename(name))) is None:
        return False
    return True

  def restore(self, checkpoint_dir):
    for name, saver in self._savers:
      full_checkpoint_dir = self._full_checkpoint_name(checkpoint_dir, name)
      logging.info("Restoring checkpoint: %s", full_checkpoint_dir)
      saver.restore(self._session, full_checkpoint_dir)

  @property
  def loss(self):
    return (self._last_critic_loss_value, self._last_pi_loss_value)

  def _add_episode_data_to_dataset(self):
    """Add episode data to the buffer."""
    info_states = [data.info_state for data in self._episode_data]
    rewards = [data.reward for data in self._episode_data]
    discount = [data.discount for data in self._episode_data]
    actions = [data.action for data in self._episode_data]

    # Calculate returns
    returns = np.array(rewards)
    for idx in reversed(range(len(rewards[:-1]))):
      returns[idx] = (
          rewards[idx] +
          discount[idx] * returns[idx + 1] * self._extra_discount)

    # Add flattened data points to dataset
    self._dataset["actions"].extend(actions)
    self._dataset["returns"].extend(returns)
    self._dataset["info_states"].extend(info_states)
    self._episode_data = []

  def _add_transition(self, time_step):
    """Adds intra-episode transition to the `_episode_data` buffer.

    Adds the transition from `self._prev_time_step` to `time_step`.

    Args:
      time_step: an instance of rl_environment.TimeStep.
    """
    assert self._prev_time_step is not None
    legal_actions = (
        self._prev_time_step.observations["legal_actions"][self.player_id])
    legal_actions_mask = np.zeros(self._num_actions)
    legal_actions_mask[legal_actions] = 1.0
    transition = Transition(
        info_state=(
            self._prev_time_step.observations["info_state"][self.player_id][:]),
        action=self._prev_action,
        reward=time_step.rewards[self.player_id],
        discount=time_step.discounts[self.player_id],
        legal_actions_mask=legal_actions_mask)

    self._episode_data.append(transition)

  def _critic_update(self):
    """Compute the Critic loss on sampled transitions & perform a critic update.

    Returns:
      The average Critic loss obtained on this batch.
    """
    # TODO(author3): illegal action handling.
    critic_loss, _ = self._session.run(
        [self._critic_loss, self._critic_learn_step],
        feed_dict={
            self._info_state_ph: self._dataset["info_states"],
            self._action_ph: self._dataset["actions"],
            self._return_ph: self._dataset["returns"],
        })
    self._last_critic_loss_value = critic_loss
    return critic_loss

  def _pi_update(self):
    """Compute the Pi loss on sampled transitions and perform a Pi update.

    Returns:
      The average Pi loss obtained on this batch.
    """
    # TODO(author3): illegal action handling.
    pi_loss, _ = self._session.run(
        [self._pi_loss, self._pi_learn_step],
        feed_dict={
            self._info_state_ph: self._dataset["info_states"],
            self._action_ph: self._dataset["actions"],
            self._return_ph: self._dataset["returns"],
        })
    self._last_pi_loss_value = pi_loss
    return pi_loss

  def get_weights(self):
    variables = [self._session.run(self._net_torso.variables)]
    variables.append(self._session.run(self._policy_logits_layer.variables))
    if self._loss_class.__name__ == "BatchA2CLoss":
      variables.append(self._session.run(self._baseline_layer.variables))
    else:
      variables.append(self._session.run(self._q_values_layer.variables))
    return variables

  def _initialize(self):
    initialization_torso = tf.group(
        *[var.initializer for var in self._net_torso.variables])
    initialization_logit = tf.group(
        *[var.initializer for var in self._policy_logits_layer.variables])
    if self._loss_class.__name__ == "BatchA2CLoss":
      initialization_baseline_or_q_val = tf.group(
          *[var.initializer for var in self._baseline_layer.variables])
    else:
      initialization_baseline_or_q_val = tf.group(
          *[var.initializer for var in self._q_values_layer.variables])
    initialization_crit_opt = tf.group(
        *[var.initializer for var in self._critic_optimizer.variables()])
    initialization_pi_opt = tf.group(
        *[var.initializer for var in self._pi_optimizer.variables()])

    self._session.run(
        tf.group(*[
            initialization_torso, initialization_logit,
            initialization_baseline_or_q_val, initialization_crit_opt,
            initialization_pi_opt
        ]))
    self._savers = [("torso", tf.train.Saver(self._net_torso.variables)),
                    ("policy_head",
                     tf.train.Saver(self._policy_logits_layer.variables))]
    if self._loss_class.__name__ == "BatchA2CLoss":
      self._savers.append(
          ("baseline", tf.train.Saver(self._baseline_layer.variables)))
    else:
      self._savers.append(
          ("q_head", tf.train.Saver(self._q_values_layer.variables)))

  def copy_with_noise(self, sigma=0.0, copy_weights=True):
    """Copies the object and perturbates its network's weights with noise.

    Args:
      sigma: gaussian dropout variance term : Multiplicative noise following
        (1+sigma*epsilon), epsilon standard gaussian variable, multiplies each
        model weight. sigma=0 means no perturbation.
      copy_weights: Boolean determining whether to copy model weights (True) or
        just model hyperparameters.

    Returns:
      Perturbated copy of the model.
    """
    _ = self._kwargs.pop("self", None)
    copied_object = PolicyGradient(**self._kwargs)

    net_torso = getattr(copied_object, "_net_torso")
    policy_logits_layer = getattr(copied_object, "_policy_logits_layer")
    if hasattr(copied_object, "_q_values_layer"):
      q_values_layer = getattr(copied_object, "_q_values_layer")
    if hasattr(copied_object, "_baseline_layer"):
      baseline_layer = getattr(copied_object, "_baseline_layer")

    if copy_weights:
      copy_mlp_weights = tf.group(*[
          va.assign(vb * (1 + sigma * tf.random.normal(vb.shape)))
          for va, vb in zip(net_torso.variables, self._net_torso.variables)
      ])
      self._session.run(copy_mlp_weights)

      copy_logit_weights = tf.group(*[
          va.assign(vb * (1 + sigma * tf.random.normal(vb.shape)))
          for va, vb in zip(policy_logits_layer.variables,
                            self._policy_logits_layer.variables)
      ])
      self._session.run(copy_logit_weights)
      if hasattr(copied_object, "_q_values_layer"):
        copy_q_value_weights = tf.group(*[
            va.assign(vb * (1 + sigma * tf.random.normal(vb.shape))) for va, vb
            in zip(q_values_layer.variables, self._q_values_layer.variables)
        ])
        self._session.run(copy_q_value_weights)
      if hasattr(copied_object, "_baseline_layer"):
        copy_baseline_weights = tf.group(*[
            va.assign(vb * (1 + sigma * tf.random.normal(vb.shape))) for va, vb
            in zip(baseline_layer.variables, self._baseline_layer.variables)
        ])
        self._session.run(copy_baseline_weights)

    for var in getattr(copied_object, "_critic_optimizer").variables():
      self._session.run(var.initializer)
    for var in getattr(copied_object, "_pi_optimizer").variables():
      self._session.run(var.initializer)

    return copied_object



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Run some analyses on some Atari data sets."""

# pylint: disable=unused-import

import sys
import time
from absl import app
from absl import flags
import numpy as np

import pyspiel
from open_spiel.python.voting import approval
from open_spiel.python.voting import base
from open_spiel.python.voting import borda
from open_spiel.python.voting import copeland
from open_spiel.python.voting import kemeny_young
from open_spiel.python.voting import maximal_lotteries
from open_spiel.python.voting import plurality
from open_spiel.python.voting import ranked_pairs
from open_spiel.python.voting import schulze
from open_spiel.python.voting import soft_condorcet_optimization as sco
from open_spiel.python.voting import stv
from open_spiel.python.voting.examples import atari_datasets

_DATASET_PATH_PREFIX = flags.DEFINE_string(
    "dataset_path_prefix", default=".", help="Where to find the dataset files")


def main(_):
  print("Loading dataset(s)...")
  dataset_filename = (_DATASET_PATH_PREFIX.value + "/" +
                      atari_datasets.RAINBOW_TABLE5)
  dataset = atari_datasets.parse_atari_table(dataset_filename)

  # If you load others, you can merge some columns from them like this:
  # dataset.add_column(dataset_ag57.get_column("random"), "random")
  # dataset.add_column(dataset_ag57.get_column("human"), "human")

  print(dataset.agent_names)
  print(dataset.game_names)
  print(f"Num agents: {len(dataset.agent_names)}")
  print(f"Num games: {len(dataset.game_names)}")

  # Alts for rainbow table 5:
  # dqn a3c ddqn prior-ddqn dueling-ddqn distrib-dqn noisy-dqn rainbow

  game_names = []
  profile = base.PreferenceProfile(alternatives=dataset.agent_names)
  for game_name, scores in dataset.table_data.items():
    profile.add_vote_from_values(scores)
    game_names.append(game_name)

  # Group up the profile and then print it to show that every vote is unique.
  profile.group()
  print(profile)

  print("Margin matrix:")
  margin_matrix = profile.margin_matrix()
  print(margin_matrix)
  print(
      "Weak Condorcet winners? "
      + f"{profile.condorcet_winner(False, margin_matrix)}"
  )
  print(
      "Strong Condorcet winner? "
      + f"{profile.condorcet_winner(True, margin_matrix)}"
  )

  voting_methods = [
      approval.ApprovalVoting(k=3),
      borda.BordaVoting(),
      copeland.CopelandVoting(),
      kemeny_young.KemenyYoungVoting(),
      maximal_lotteries.MaximalLotteriesVoting(iterative=True),
      plurality.PluralityVoting(),
      ranked_pairs.RankedPairsVoting(),
      schulze.SchulzeVoting(),
      stv.STVVoting(num_winners=3),
  ]
  for method in voting_methods:
    print("")
    print(method.name())
    outcome = method.run_election(profile)
    print(outcome.pretty_table_string())

  print("Soft Condorcet Optimization (Python):")
  py_sco_solver = sco.SoftCondorcetOptimizer(
      profile,
      batch_size=4,
      rating_lower_bound=-100.0,
      rating_upper_bound=100.0,
      temperature=1,
  )
  start_time = time.time()
  ratings, ranking = py_sco_solver.run_solver(10000, learning_rate=0.01)
  end_time = time.time()
  print(f"Time taken: {end_time - start_time}")
  alt_idx = profile.alternatives_dict
  for alt in ranking:
    print(f"  {alt}: {ratings[alt_idx[alt]]}")

  print("Soft Condorcet Optimization Sigmoid (C++):")
  cpp_sco_solver = pyspiel.sco.SoftCondorcetOptimizer(
      profile.to_list_of_tuples(),
      rating_lower_bound=-100.0,
      rating_upper_bound=100.0,
      batch_size=4,
      temperature=1,
      rng_seed=0,
  )
  start_time = time.time()
  cpp_sco_solver.run_solver(10000, learning_rate=0.01)
  end_time = time.time()
  print(f"Time taken: {end_time - start_time}")
  ratings_dict = cpp_sco_solver.ratings()
  for alt in ranking:
    print(f"  {alt}: {ratings_dict[alt]}")

  print("Soft Condorcet Optimization Fenchel-Young (C++):")
  cpp_fy_solver = pyspiel.sco.FenchelYoungOptimizer(
      profile.to_list_of_tuples(),
      rating_lower_bound=-100.0,
      rating_upper_bound=100.0,
      batch_size=4,
      temperature=1,
      rng_seed=0,
  )
  start_time = time.time()
  cpp_fy_solver.run_solver(10000, learning_rate=0.01)
  end_time = time.time()
  print(f"Time taken: {end_time - start_time}")
  ratings_dict = cpp_fy_solver.ratings()
  for alt in ranking:
    print(f"  {alt}: {ratings_dict[alt]}")


if __name__ == "__main__":
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Useful sequence form functions used in the MMD implementation."""

import numpy as np
from open_spiel.python import policy


_DELIMITER = " -=- "


def get_isa_key(info_state, action):
  return info_state + _DELIMITER + str(action)


def get_action_from_key(isa_key):
  _, action_str = isa_key.split(_DELIMITER)
  return int(action_str)


def get_infostate_from_key(isa_key, player):
  assert not is_root(isa_key, player), "Cannot use this method for root nodes."
  infostate, _ = isa_key.split(_DELIMITER)
  return infostate


def is_root(key, player):
  empty_is_key = f"***EMPTY_INFOSET_P{player}***"
  empty_isa_key = f"***EMPTY_INFOSET_ACTION_P{player}***"
  return True if key in [empty_is_key, empty_isa_key] else False


def construct_vars(game):
  """Construct useful sequence from variables from game.

  Args:
      game: The spiel game to solve (must be zero-sum, sequential, and have
        chance node of deterministic or explicit stochastic).

  Returns:
      An 8 tuple of sequence form variables from _construct_vars by
      recursively
      traversing the game tree.

  """

  initial_state = game.new_initial_state()
  npl = game.num_players()

  empty_is_keys = [f"***EMPTY_INFOSET_P{player}***" for player in range(npl)]
  empty_isa_keys = [
      f"***EMPTY_INFOSET_ACTION_P{player}***" for player in range(npl)
  ]

  # initialize variables
  infosets = [{empty_is_keys[p]: 0} for p in range(npl)]
  infoset_actions_to_seq = [{empty_isa_keys[p]: 0} for p in range(npl)]
  infoset_action_maps = [
      {empty_is_keys[p]: [empty_isa_keys[p]]} for p in range(npl)
  ]

  # infoset_action_maps = [{}, {}]
  payoff_dict = dict()

  infoset_parent_map = [{empty_isa_keys[p]: None} for p in range(npl)]
  infoset_actions_children = [{empty_isa_keys[p]: []} for p in range(npl)]

  _construct_vars(initial_state, infosets, infoset_actions_to_seq,
                  infoset_action_maps, infoset_parent_map, 1.0,
                  empty_is_keys[:], empty_isa_keys[:],
                  payoff_dict, infoset_actions_children)

  payoff_mat = _construct_numpy_vars(payoff_dict, infoset_actions_to_seq)
  return (infosets, infoset_actions_to_seq,
          infoset_action_maps, infoset_parent_map,
          payoff_mat, infoset_actions_children)


def uniform_random_seq(game, infoset_actions_to_seq):
  """Generate uniform random sequence.

  The sequence generated is equivalent to a uniform random tabular policy.

  Args:
      game: the spiel game to solve (must be zero-sum, sequential, and have
        chance mode of deterministic or explicit stochastic).
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.

  Returns:
      A list of NumPy arrays, one for each player.
  """
  policies = policy.TabularPolicy(game)
  initial_state = game.new_initial_state()
  sequences = [
      np.ones(len(infoset_actions_to_seq[i])) for i in range(game.num_players())
  ]
  _policy_to_sequence(
      initial_state,
      policies,
      sequences,
      infoset_actions_to_seq,
      [1 for _ in range(game.num_players())],
  )
  return sequences


def _construct_vars(state, infosets, infoset_actions_to_seq,
                    infoset_action_maps, infoset_parent_map, chance_reach,
                    parent_is_keys, parent_isa_keys, payoff_dict,
                    infoset_actions_children):
  """Recursively builds maps and the sequence form payoff matrix.

  Args:
      state: pyspiel (OpenSpiel) state
      infosets: a list of dicts, one per player, that maps infostate to an id.
        The dicts are filled by this function and should initially only
        contain root values.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id. The dicts are filled by
        this function and should initially only contain the root values.
      infoset_action_maps: a list of dicts, one per player, that maps each
        info_state to a list of (infostate, action) string.
      infoset_parent_map: a list of dicts, one per player, that maps each
        info_state to an (infostate, action) string.
      chance_reach: the contribution of chance's reach probability (should
        start at 1).
      parent_is_keys: a list of parent information state keys for this state
      parent_isa_keys: a list of parent (infostate, action) keys
      payoff_dict: a dict that maps sequences of players' (infostate, action)
        tuples, e.g., ((infostate, action), ...) to the chance weighted reward.
      infoset_actions_children: a list of dicts, one for each player, mapping
        (infostate, action) keys to reachable infostates for each player
  """

  if state.is_terminal():
    returns = state.returns()
    idx = tuple(parent_isa_keys_i for parent_isa_keys_i in parent_isa_keys)
    payoff_dict.setdefault(idx, 0)
    payoff_dict[idx] += np.asarray(returns) * chance_reach
    return

  if state.is_chance_node():
    for action, prob in state.chance_outcomes():
      new_state = state.child(action)
      _construct_vars(new_state, infosets, infoset_actions_to_seq,
                      infoset_action_maps, infoset_parent_map,
                      prob * chance_reach, parent_is_keys, parent_isa_keys,
                      payoff_dict, infoset_actions_children)
    return

  player = state.current_player()
  info_state = state.information_state_string(player)
  legal_actions = state.legal_actions(player)

  # Add to the infostate maps
  if info_state not in infosets[player]:
    infosets[player][info_state] = len(infosets[player])
  if info_state not in infoset_action_maps[player]:
    infoset_action_maps[player][info_state] = []

  # Add to infoset to parent infoset action map
  if info_state not in infoset_parent_map[player]:
    infoset_parent_map[player][info_state] = parent_isa_keys[player]

  # add as child to parent
  if parent_isa_keys[player] in infoset_actions_children[player]:
    if info_state not in infoset_actions_children[player][
        parent_isa_keys[player]]:
      infoset_actions_children[player][parent_isa_keys[player]].append(
          info_state)
  else:
    infoset_actions_children[player][parent_isa_keys[player]] = [info_state]

  new_parent_is_keys = parent_is_keys[:]
  new_parent_is_keys[player] = info_state

  for action in legal_actions:
    isa_key = get_isa_key(info_state, action)
    if isa_key not in infoset_actions_to_seq[player]:
      infoset_actions_to_seq[player][isa_key] = len(
          infoset_actions_to_seq[player])
    if isa_key not in infoset_action_maps[player][info_state]:
      infoset_action_maps[player][info_state].append(isa_key)

    new_parent_isa_keys = parent_isa_keys[:]
    new_parent_isa_keys[player] = isa_key
    new_state = state.child(action)
    _construct_vars(new_state, infosets, infoset_actions_to_seq,
                    infoset_action_maps, infoset_parent_map, chance_reach,
                    new_parent_is_keys, new_parent_isa_keys, payoff_dict,
                    infoset_actions_children)


def _construct_numpy_vars(payoff_dict, infoset_actions_to_seq):
  """Convert sequence form payoff dict to numpy array.

  Args:
      payoff_dict: a dict that maps sequences of players' (infostate, action)
        tuples, e.g., ((infostate, action), ...) to the chance weighted reward.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.

  Returns:
      A numpy array corresponding to the chance weighted rewards
      i.e. the sequence form payoff tensor.

  """
  npl = len(infoset_actions_to_seq)
  pls = range(npl)  # player list
  sequence_sizes = tuple(len(infoset_actions_to_seq[i]) for i in pls)
  payoff_tensor = np.zeros((npl,) + sequence_sizes)
  for player_isa_seqs, payoffs in payoff_dict.items():
    idx = tuple(infoset_actions_to_seq[i][player_isa_seqs[i]] for i in pls)
    payoff_tensor[(slice(None),) + idx] = np.asarray(payoffs)
  return payoff_tensor


def construct_constraint_vars(infoset_parent_map, infoset_actions_to_seq,
                              infoset_action_maps):
  """Construct useful sequence form variables from game.

  Args:
      infoset_parent_map: a list of dicts, one per player, that maps each
        info_state to an (infostate, action) string.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.
      infoset_action_maps: a list of dicts, one per player, that maps each
        info_state to a list of (infostate, action) string.

  Returns:
      A dict mapping player to a tuple containing a numpy array of coefficients,
      each of dimension # of player sequences, as well as a sparse vector
      containing the constants, i.e., dict[player] = (A, b) as in Ax = b.
  """
  npl = len(infoset_actions_to_seq)
  constraint_dict = {}

  for player in range(npl):
    num_seqs = len(infoset_actions_to_seq[player].values())

    root_con = np.zeros(num_seqs)
    root_con[0] = 1.0
    constraints = [root_con]

    for info_state in infoset_action_maps[player]:
      if is_root(info_state, player):
        continue

      parent_isa_key = infoset_parent_map[player][info_state]
      parent_seq_id = infoset_actions_to_seq[player][parent_isa_key]

      # seq ids for children
      children_isa_keys = infoset_action_maps[player][info_state]
      children_seq_ids = [
          infoset_actions_to_seq[player][isa_key]
          for isa_key in children_isa_keys
      ]

      constraint = np.zeros(num_seqs)
      constraint[parent_seq_id] = -1.0
      constraint[children_seq_ids] = 1.0
      constraints.append(constraint)

    constant = np.zeros(len(constraints))
    constant[0] = 1.0
    constraint_dict[player] = (np.stack(constraints), constant)

  return constraint_dict


def sequence_to_policy(sequences, game, infoset_actions_to_seq,
                       infoset_action_maps):
  """Convert sequence form policies to the realization-equivalent tabular ones.

  Args:
      sequences: list of two sequence form policies, one for each player.
      game: a spiel game with two players.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.
      infoset_action_maps: a list of dicts, one per player, that maps each
        info_state to a list of (infostate, action) string.

  Returns:
      A TabularPolicy object.
  """

  policies = policy.TabularPolicy(game)
  for player in range(2):
    for info_state in infoset_action_maps[player]:
      if is_root(info_state, player):
        continue

      state_policy = policies.policy_for_key(info_state)
      total_weight = 0
      num_actions = 0

      for isa_key in infoset_action_maps[player][info_state]:
        total_weight += sequences[player][infoset_actions_to_seq[player]
                                          [isa_key]]
        num_actions += 1

      unif_pr = 1.0 / num_actions
      for isa_key in infoset_action_maps[player][info_state]:
        rel_weight = sequences[player][infoset_actions_to_seq[player][isa_key]]
        _, action_str = isa_key.split(_DELIMITER)
        action = int(action_str)
        pr_action = rel_weight / total_weight if total_weight > 0 else unif_pr
        state_policy[action] = pr_action
  return policies


def policy_to_sequence(game, policies, infoset_actions_to_seq):
  """Converts a TabularPolicy object for a two-player game.

  The converted policy is its realization-equivalent sequence form one.

  Args:
      game: a two-player open spiel game.
      policies: a TabularPolicy object.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.

  Returns:
      A list of numpy arrays, one for each player.
  """
  initial_state = game.new_initial_state()
  sequences = [
      np.ones(len(infoset_actions_to_seq[0])),
      np.ones(len(infoset_actions_to_seq[1]))
  ]
  _policy_to_sequence(initial_state, policies, sequences,
                      infoset_actions_to_seq, [1, 1])
  return sequences


def _policy_to_sequence(state, policies, sequences, infoset_actions_to_seq,
                        parent_seq_val):
  """Converts a TabularPolicy object to its equivalent sequence form.

  This method modifies the sequences inplace and should not be called directly.

  Args:
      state: an openspiel state.
      policies: a TabularPolicy object.
      sequences: list of numpy arrays to be modified.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.
      parent_seq_val: list of parent sequence values, this method should be
        called with initial value of [1,1].
  """

  if state.is_terminal():
    return

  if state.is_chance_node():
    for action, _ in state.chance_outcomes():
      new_state = state.child(action)
      _policy_to_sequence(new_state, policies, sequences,
                          infoset_actions_to_seq, parent_seq_val)
    return

  player = state.current_player()
  info_state = state.information_state_string(player)
  legal_actions = state.legal_actions(player)
  state_policy = policies.policy_for_key(info_state)
  for action in legal_actions:
    isa_key = get_isa_key(info_state, action)
    # update sequence form
    sequences[player][infoset_actions_to_seq[player]
                      [isa_key]] = parent_seq_val[player] * state_policy[action]
    new_parent_seq_val = parent_seq_val[:]
    new_parent_seq_val[player] = sequences[player][
        infoset_actions_to_seq[player][isa_key]]
    new_state = state.child(action)
    _policy_to_sequence(new_state, policies, sequences, infoset_actions_to_seq,
                        new_parent_seq_val)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A few prompts for extracting the winner of a debate.
"""

from open_spiel.python.games.chat_games.envs.observations import summary_debate
from open_spiel.python.games.chat_games.envs.payoffs import utils as payoff_utils
from open_spiel.python.games.chat_games.envs.utils import text

MIN_PAYOFF = 0
MAX_PAYOFF = 1
PAYOFF_PROMPT_a = [
    'You are an assistant designed to determine the winner of a debate. ' +
    'Read the arguments by each side in the debate. Then determine if ' +
    'the player won the debate. Here is an example enclosed by "&".']

PAYOFF_PROMPT_b = [
    # '&' * 50,
    'Bob argues that Tom Brady is the GOAT and coach Bill Bellichick is a ' +
    'genius.',
    'Alice argues that Tom Brady is no longer on the Patriots and the ' +
    'Patriots were 8-9 last season. They lost more games than they won. They ' +
    'are no longer the powerhouse they used to be.',
    'Bob makes an argument based on stale information. Alice acutely points ' +
    'this out and provides more current evidence that supports the negative ' +
    'of Bobs argument. Therefore, Bob loses the debate.',
    'Value for Bob: 0.',
    'Value for Alice: 1.',
    '&' * 50,
    'Now determine the winner of the following debate.',
    '{m}',
    '%' * 50,
    'Payoff for {p} ONLY: ']

PAYOFF_PROMPT = ('\n\n'.join(text.wrap(PAYOFF_PROMPT_a)) + '\n\n' + '&' * 50 +
                 '\n\nDebate Topic: The New England Patriots are the best ' +
                 'NFL team in 2023.\n\n' +
                 '\n\n'.join(text.wrap(PAYOFF_PROMPT_b)))

PAYOFF_OBS_TRANS_PREFIX = summary_debate.PREFIX

PAYOFF_OBS_TRANS_POSTFIX = summary_debate.POSTFIX

PAYOFF = payoff_utils.Payoff(PAYOFF_PROMPT,
                             MIN_PAYOFF,
                             MAX_PAYOFF,
                             PAYOFF_OBS_TRANS_PREFIX,
                             PAYOFF_OBS_TRANS_POSTFIX)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python Kuhn Poker."""

from absl.testing import absltest
import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import sequence_form_lp
from open_spiel.python.algorithms.get_all_states import get_all_states
from open_spiel.python.games import kuhn_poker  # pylint: disable=unused-import
from open_spiel.python.observation import make_observation
import pyspiel


class KuhnPokerTest(absltest.TestCase):

  def test_game_from_cc(self):
    """Runs our standard game tests, checking API consistency."""
    game = pyspiel.load_game("python_kuhn_poker")
    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)

  def test_consistent(self):
    """Checks the Python and C++ game implementations are the same."""
    py_game = pyspiel.load_game("python_kuhn_poker")
    cc_game = pyspiel.load_game("kuhn_poker")
    obs_types = [None, pyspiel.IIGObservationType(perfect_recall=True)]
    py_observations = [make_observation(py_game, o) for o in obs_types]
    cc_observations = [make_observation(cc_game, o) for o in obs_types]
    py_states = get_all_states(py_game)
    cc_states = get_all_states(cc_game)
    self.assertCountEqual(list(cc_states), list(py_states))
    for key, cc_state in cc_states.items():
      py_state = py_states[key]
      np.testing.assert_array_equal(py_state.history(), cc_state.history())
      np.testing.assert_array_equal(py_state.returns(), cc_state.returns())
      for py_obs, cc_obs in zip(py_observations, cc_observations):
        for player in (0, 1):
          py_obs.set_from(py_state, player)
          cc_obs.set_from(cc_state, player)
          np.testing.assert_array_equal(py_obs.tensor, cc_obs.tensor)

  def test_nash_value_sequence_form_lp(self):
    """Checks Nash value using a Python sequence form LP solver."""
    game = pyspiel.load_game("python_kuhn_poker")
    val1, val2, _, _ = sequence_form_lp.solve_zero_sum_game(game)
    # value from Kuhn 1950 or https://en.wikipedia.org/wiki/Kuhn_poker
    self.assertAlmostEqual(val1, -1 / 18)
    self.assertAlmostEqual(val2, +1 / 18)

  def test_exploitability_uniform_random_py(self):
    """Checks the exploitability of the uniform random policy using Python."""
    # NashConv of uniform random test_policy from (found on Google books):
    # https://link.springer.com/chapter/10.1007/978-3-319-75931-9_5
    game = pyspiel.load_game("python_kuhn_poker")
    test_policy = policy.UniformRandomPolicy(game)
    expected_nash_conv = 11 / 12
    self.assertAlmostEqual(
        exploitability.exploitability(game, test_policy),
        expected_nash_conv / 2)

  def test_exploitability_uniform_random_cc(self):
    """Checks the exploitability of the uniform random policy using C++."""
    game = pyspiel.load_game("python_kuhn_poker")
    test_policy = pyspiel.UniformRandomPolicy(game)
    expected_nash_conv = 11 / 12
    self.assertAlmostEqual(
        pyspiel.exploitability(game, test_policy), expected_nash_conv / 2)

  def test_cfr_cc(self):
    """Runs a C++ CFR algorithm on the game."""
    game = pyspiel.load_game("python_kuhn_poker")
    unused_results = pyspiel.CFRSolver(game)


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import numpy as np
from open_spiel.python.voting import base
from open_spiel.python.voting import maximal_lotteries


class MaximalLotteriesTest(parameterized.TestCase):
  @parameterized.named_parameters(("iterative", True), ("non-iterative", False))
  def test_stv_records_number(self, iterate):
    method = maximal_lotteries.MaximalLotteriesVoting(iterative=iterate)
    self.assertEqual(
        method.name(), f"maximal_lotteries(iterative={iterate})"
    )

  def test_maximal_lotteries_basic_run(self):
    # "a" is a dominant strategy of the margin game, so it should be chosen with
    # probablity 1.
    votes = [["a", "b", "c"], ["a", "c", "b"], ["b", "a", "c"]]
    profile = base.PreferenceProfile(votes=votes)
    method = maximal_lotteries.MaximalLotteriesVoting(iterative=False)
    outcome = method.run_election(profile)
    with self.subTest("Top-rank the condorcet winner"):
      self.assertEqual(outcome.ranking[0], "a")
    with self.subTest("Check extreme scores"):
      self.assertAlmostEqual(outcome.scores[0], 1.0)
      self.assertAlmostEqual(outcome.scores[1], 0.0)
      self.assertAlmostEqual(outcome.scores[2], 0.0)

  def test_maximal_lotteries_basic_iterative(self):
    votes = [["a", "b", "c"], ["a", "c", "b"], ["b", "a", "c"]]
    profile = base.PreferenceProfile(votes=votes)
    # "a" is a dominant strategy, so in the iterative version it should be
    # chosen first, leading to a new matrix with the first row and column
    # deleted. This then means that "b" is dominant in the subgame.
    expected_margin_matrix = np.array([
        [0, 1, 3],
        [-1, 0, 1],
        [-3, -1, 0]])
    with self.subTest("Check margin matrix"):
      self.assertTrue(np.array_equal(profile.margin_matrix(),
                                     expected_margin_matrix))
    method = maximal_lotteries.MaximalLotteriesVoting(iterative=True)
    outcome = method.run_election(profile)
    with self.subTest("Check ranking"):
      self.assertListEqual(outcome.ranking, ["a", "b", "c"])
    with self.subTest("Check scores"):
      self.assertAlmostEqual(outcome.scores[0], 3.0)
      self.assertAlmostEqual(outcome.scores[1], 2.0)
      self.assertAlmostEqual(outcome.scores[2], 1.0)

  def test_maximal_lotteries_cycle(self):
    # Cyclical profile leads to a Rock, Paper, Scissors margin game.
    votes = [["a", "b", "c"], ["b", "c", "a"], ["c", "a", "b"]]
    profile = base.PreferenceProfile(votes=votes)
    method = maximal_lotteries.MaximalLotteriesVoting()
    outcome = method.run_election(profile)
    with self.subTest("Check prob 1/3"):
      self.assertAlmostEqual(outcome.scores[0], 1.0 / 3.0)
    with self.subTest("Check uniform"):
      self.assertAlmostEqual(outcome.scores[0], outcome.scores[1])
      self.assertAlmostEqual(outcome.scores[1], outcome.scores[2])

  def test_maximal_lotteries_iterative_cycle(self):
    # Cyclical profile leads to a Rock, Paper, Scissors margin game.
    # Iterative maximal lotteries should yield the same result as the
    # non-iterative version.
    votes = [["a", "b", "c"], ["b", "c", "a"], ["c", "a", "b"]]
    profile = base.PreferenceProfile(votes=votes)
    method = maximal_lotteries.MaximalLotteriesVoting(iterative=True)
    outcome = method.run_election(profile)
    with self.subTest("Check prob 1/3"):
      self.assertAlmostEqual(outcome.scores[0], 1.0 / 3.0)
    with self.subTest("Check uniform"):
      self.assertAlmostEqual(outcome.scores[0], outcome.scores[1])
      self.assertAlmostEqual(outcome.scores[1], outcome.scores[2])


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for loading Atari data."""

import logging
from typing import Dict, List
import numpy as np


RAINBOW_TABLE5 = "atari_rainbow_table5.txt"
RAINBOW_TABLE6 = "atari_rainbow_table6.txt"
AGENT57_TABLE = "atari_agent57_table.txt"
MUESLI_TABLE11 = "atari_muesli_table11.txt"


class DataSet:
  """A DataSet container for Atari tables."""

  def __init__(
      self,
      agent_names: List[str],
      game_names: List[str],
      table_data: Dict[str, List[float]],
  ):
    self.agent_names = agent_names
    self.game_names = game_names
    self.table_data = table_data

  def get_column(self, agent_name: str) -> Dict[str, float]:
    column_dict = {}
    agent_idx = self.agent_names.index(agent_name)
    assert 0 <= agent_idx < len(self.agent_names)
    for game_name, scores in self.table_data.items():
      column_dict[game_name] = scores[agent_idx]
    return column_dict

  def delete_column(self, agent_name: str):
    agent_idx = self.agent_names.index(agent_name)
    assert 0 <= agent_idx < len(self.agent_names)
    del self.agent_names[agent_idx]
    for game_name in self.game_names:
      del self.table_data[game_name][agent_idx]

  def delete_game(self, game_name: str):
    assert game_name in self.game_names
    self.game_names.remove(game_name)
    del self.table_data[game_name]

  def add_column(self, column, agent_name):
    """Add a column.

    Args:
        column: a dictionary of game_name -> score,
        agent_name: name for the new agent.

    Note: beware! This can delete rows within this data set, in order to keep
    data complete, i.e. it deletes rows if you don't have this agent's score for
    that game.
    """
    self.agent_names.append(agent_name)
    game_names_copy = self.game_names[:]
    for game_name in game_names_copy:
      if game_name not in column:
        logging.warning("Warning: deleting game {%s}", game_name)
        self.delete_game(game_name)
      else:
        self.table_data[game_name].append(column[game_name])

  def to_task_by_agent_matrix(self) -> np.ndarray:
    num_tasks = len(self.game_names)
    num_agents = len(self.agent_names)
    mat = np.zeros(shape=(num_tasks, num_agents))
    i = 0
    for game_name in self.game_names:
      mat[i] = np.asarray(self.table_data[game_name])
      i += 1
    return mat


def parse_value(val_str: str) -> float:
  """Parse a numerical value from string, dropping ± part."""
  val_str = val_str.replace(",", "")
  val_str = val_str.split("±")[0]
  return float(val_str)


def parse_values(string_values_list: List[str]) -> List[float]:
  """Turn a list of strings into a list of floats."""
  return [parse_value(val) for val in string_values_list]


def delete_agent(dataset: DataSet, agent: str):
  idx = dataset.agent_names.index(agent)
  assert 0 <= idx < len(dataset.agent_names)
  del dataset.agent_names[idx]
  for key in dataset.table_data.keys():
    del dataset.table_data[key][idx]


def make_subset(dataset: DataSet, agent_subset: List[str]):
  for agent in dataset.agent_names:
    if agent not in agent_subset:
      delete_agent(dataset, agent)


def parse_atari_table(filename: str) -> DataSet:
  """Parse an Atari data file.

  The files are created by copy/paste from the papers.

  Args:
    filename: the file that contains the dataset.

  Returns:
      a DataSet object referring to the Atari data.
  """
  with open(filename, "r") as f:
    string_data = f.read()

  # First line is a comment
  # Second line format is column descriptions, e.g.:
  # "# game <agent1 name> <agent2 name> ..."
  # Rest of the lines are copy/paste from the paper tables.
  lines = string_data.split("\n")
  assert lines[1].startswith("# game ")
  agent_names = lines[1].split()[2:]
  num_agents = len(agent_names)
  game_names = []
  table_data = {}
  for i in range(2, len(lines)):
    if lines[i].strip():
      parts = lines[i].split()
      game_name = parts[0]
      game_names.append(game_name)
      str_scores = parts[1:]
      assert len(str_scores) == num_agents, f"Error line: {lines[i]}"
      scores = parse_values(str_scores)
      table_data[game_name] = scores
  return DataSet(agent_names, game_names, table_data)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module implements a generic environment for iterated normal form games.

It does so wuth automatic vectorization. Along with the environment, it also
provides pre-defined factory functions for common games such as the iterated
prisoners dilemma and the iterated matching pennies.
"""

import numpy as np
from pyspiel import PlayerId

from open_spiel.python.rl_environment import Environment
from open_spiel.python.rl_environment import StepType
from open_spiel.python.rl_environment import TimeStep


class IteratedMatrixGame(Environment):
  """Environment for iterated normal form games.

  Supports automatic vectorization.
  """

  def __init__(
      self,
      payoff_matrix: np.ndarray,
      iterations: int,
      batch_size=1,
      include_remaining_iterations=True,
  ):
    # pylint: disable=super-init-not-called
    self._payoff_matrix = np.array(payoff_matrix, dtype=np.float32)
    self._iterations = iterations
    self._num_players = payoff_matrix.ndim - 1
    self._batch_size = batch_size
    self._include_remaining_iterations = include_remaining_iterations
    self._t = 0
    self._actions = np.arange(
        np.prod(self.action_spec()['num_actions'])
    ).reshape(*[payoff_matrix.shape[p] for p in range(self._num_players)])

  def one_hot(self, x, n):
    return np.eye(n)[x]

  @property
  def num_players(self):
    return self._num_players

  def observation_spec(self):
    info_state_spec, legal_actions_spec = [], []
    for i in range(self._num_players):
      num_actions = np.prod(self._payoff_matrix.shape[:-1]) + 1
      if self._include_remaining_iterations:
        num_actions += 1
      info_state_spec.append([num_actions])
      legal_actions_spec.append(self._payoff_matrix.shape[i])
    return {
        'info_state': tuple(info_state_spec),
        'legal_actions': tuple(legal_actions_spec),
        'current_player': (),
    }

  def action_spec(self):
    num_actions, mins, maxs = [], [], []
    for i in range(self._num_players):
      num_actions.append(self._payoff_matrix.shape[i])
      mins.append(0)
      maxs.append(self._payoff_matrix.shape[i] - 1)

    return {
        'num_actions': tuple(num_actions),
        'min': tuple(mins),
        'max': tuple(maxs),
        'dtype': int,
    }

  def step(self, actions: np.ndarray):
    if actions.ndim == 1:
      actions = actions[None, :]
    payoffs = self._payoff_matrix[tuple(actions.T)]
    s1 = self.one_hot(
        self._actions[tuple(actions.T)] + 1, n=np.max(self._actions) + 2
    )
    s2 = self.one_hot(
        self._actions[tuple(actions[..., ::-1].T)] + 1,
        n=np.max(self._actions) + 2,
    )
    rewards = [
        np.squeeze(p)
        for p in np.split(
            payoffs, indices_or_sections=self._num_players, axis=1
        )
    ]
    discounts = [np.ones_like(r) for r in rewards]
    if self._t == self._iterations - 1:
      step_type = StepType.LAST
    else:
      step_type = StepType.MID
    self._t += 1
    remaining_iters = float((self._iterations - self._t)) / self._iterations

    info_state = [s1, s2]
    if self._include_remaining_iterations:
      info_state = np.concatenate(
          [
              info_state,
              np.full((self._batch_size, 1), fill_value=remaining_iters),
          ],
          axis=-1,
      )

    legal_actions = self._get_legal_actions()
    return TimeStep(
        observations={
            'info_state': info_state,
            'legal_actions': legal_actions,
            'batch_size': actions.shape[0],
            'current_player': PlayerId.SIMULTANEOUS,
        },
        rewards=rewards,
        discounts=discounts,
        step_type=step_type,
    )

  def _get_legal_actions(self):
    legal_actions = []
    for p in range(self.num_players):
      actions = np.arange(self.action_spec()['num_actions'][p])
      legal_actions.append([actions] * self._batch_size)
    return np.array(legal_actions)

  def reset(self):
    self._t = 0
    info_state = np.zeros((
        self.num_players,
        self._batch_size,
        *self.observation_spec()['info_state'][0],
    ))
    info_state[..., 0] = 1.0
    if self._include_remaining_iterations:
      info_state[..., -1] = 1.0
    rewards = np.squeeze(np.zeros((self.num_players, self._batch_size)))
    discounts = np.squeeze(np.ones((self.num_players, self._batch_size)))
    return TimeStep(
        observations={
            'info_state': [
                np.squeeze(s).astype(np.float32) for s in info_state
            ],
            'legal_actions': self._get_legal_actions(),
            'batch_size': self._batch_size,
            'current_player': PlayerId.SIMULTANEOUS,
        },
        rewards=[np.squeeze(a).astype(np.float32) for a in rewards],
        discounts=[np.squeeze(a).astype(np.float32) for a in discounts],
        step_type=StepType.FIRST,
    )


def IteratedPrisonersDilemma(iterations: int, batch_size=1):
  return IteratedMatrixGame(
      payoff_matrix=np.array([[[-1, -1], [-3, 0]], [[0, -3], [-2, -2]]]),
      iterations=iterations,
      batch_size=batch_size,
      include_remaining_iterations=False,
  )


def IteratedMatchingPennies(iterations: int, batch_size=1):
  return IteratedMatrixGame(
      payoff_matrix=np.array([[[1, -1], [-1, 1]], [[-1, 1], [1, -1]]]),
      iterations=iterations,
      batch_size=batch_size,
      include_remaining_iterations=False,
  )



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mean-Field PSRO.

As implemented in Muller et al., 2021, https://arxiv.org/abs/2111.08350
"""

from open_spiel.python import policy as policy_std
from open_spiel.python.algorithms import get_all_states
from open_spiel.python.mfg.algorithms import correlated_equilibrium
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import greedy_policy


def dict_equal(dic1, dic2):
  return all([dic1[a] == dic2[a] for a in dic1]) and all(
      [dic1[a] == dic2[a] for a in dic2]
  )


def equal_policies(pol1, pol2, all_states):
  assert isinstance(pol1, greedy_policy.GreedyPolicy)
  equal = True
  for state_key in all_states:
    state = all_states[state_key]
    try:
      equal = equal and dict_equal(pol1(state), pol2(state))
    except KeyError:
      equal = False
    except ValueError:
      continue
  return equal


def filter_policies(policies, new_policies, all_states):
  all_policies = policies
  no_novelty = True
  for new_policy in new_policies:
    if all([
        not equal_policies(new_policy, policy, all_states)
        for policy in all_policies
    ]):
      all_policies.append(new_policy)
      no_novelty = False
  return all_policies, no_novelty


class MeanFieldPSRO:
  """Mean-Field PSRO."""

  def __init__(
      self,
      game,
      regret_minimizer,
      regret_steps_per_step,
      best_responder=correlated_equilibrium.cce_br,
      filter_new_policies=False,
      increase_precision_when_done_early=False,
  ):
    self._game = game
    self._regret_minimizer = regret_minimizer
    self._regret_steps_per_step = regret_steps_per_step

    self._filter_new_policies = filter_new_policies
    self._increase_precision_when_done_early = (
        increase_precision_when_done_early
    )

    self._best_responder = best_responder

    self._nus = [[1.0]]
    self._policies = [policy_std.UniformRandomPolicy(self._game)]
    self._mus = [distribution.DistributionPolicy(game, self._policies[0])]
    self._weights = [1.0]

    self._all_states = None
    if self._filter_new_policies:
      self._all_states = get_all_states.get_all_states(game)

  def step(self):
    """Does a best-response step."""
    rewards = self._regret_minimizer.get_rewards()

    print("Computing best response.")
    new_policies, gap_value = self._best_responder(
        self._game, self._policies, self._weights, self._mus, self._nus, rewards
    )

    no_novelty = False
    if self._filter_new_policies:
      print("Filtering best responses")
      self._policies, no_novelty = filter_policies(
          self._policies, new_policies, self._all_states
      )
    else:
      self._policies = self._policies + new_policies

    if no_novelty:
      print("No new policy added, PSRO has terminated.")
      if self._increase_precision_when_done_early:
        print("Increasing precision")
        self._regret_minimizer.increase_precision_x_fold(2.0)
        self._regret_steps_per_step *= 2
        self._regret_minimizer.restart()
        self._regret_minimizer.step_for(self._regret_steps_per_step)
    else:
      print("Minimizing regret")
      self._regret_minimizer.reset(self._policies)
      self._regret_minimizer.step_for(self._regret_steps_per_step)

    average_regret = self._regret_minimizer.compute_average_regret()
    print("Average Regret : {}".format(average_regret))

    self._mus, self._weights = self._regret_minimizer.get_mus_and_weights()
    self._nus = self._regret_minimizer.get_nus()
    return average_regret, gap_value

  def get_equilibrium(self):
    return self._policies, self._nus, self._mus, self._weights



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Some basic coalitional games.

Many of these are taken from examples in the "Computational Aspects of
Cooperative Game Theory" text book by Georgios Chalkiadakis, Edith Elkind, and
Michael Wooldridge.
"""

from typing import Dict, Tuple

import numpy as np

from open_spiel.python.coalitional_games import coalitional_game


class IceCreamGame(coalitional_game.CoalitionalGame):
  """Example 2.2 from CACGT book by Chalkiadakis, Elkind, and Wooldridge."""

  def __init__(self):
    super().__init__(num_players=3)

  def coalition_value(self, coalition: np.ndarray) -> float:
    """Encodes the payoffs."""
    # indices ordered as C M P
    if coalition.sum() < 2:
      return 0.0
    elif coalition[0] == 1 and coalition[1] == 1 and coalition[2] == 0:
      # {C, M}
      return 500.0
    elif coalition[0] == 1 and coalition[1] == 0 and coalition[2] == 1:
      # {C, P}
      return 500.0
    elif coalition[0] == 0 and coalition[1] == 1 and coalition[2] == 1:
      # {M, P}
      return 750.0
    elif coalition.sum() == 3:
      return 1000.0
    else:
      raise RuntimeError("Invalid coalition")


class TabularGame(coalitional_game.CoalitionalGame):
  """A game represented by a table of values."""

  def __init__(self, table: Dict[Tuple[int, ...], float]):
    super().__init__(num_players=-1)  # set num players to -1 for now
    for key in table:
      if self._num_players < 0:
        self._num_players = len(key)
      else:
        assert len(key) == self._num_players
    assert self._num_players >= 1
    self._table = table

  def coalition_value(self, coalition: np.ndarray) -> float:
    return self._table[tuple(coalition)]




# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for best_response_value."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python import policy
from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import best_response_value
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.games import crowd_modelling  # pylint: disable=unused-import
import pyspiel


class BestResponseTest(parameterized.TestCase):

  @parameterized.named_parameters(('python', 'python_mfg_crowd_modelling'),
                                  ('cpp', 'mfg_crowd_modelling'))
  def test_best_response(self, name):
    """Checks if the value of a policy computation works."""
    game = pyspiel.load_game(name)
    uniform_policy = policy.UniformRandomPolicy(game)
    dist = distribution.DistributionPolicy(game, uniform_policy)
    br_value = best_response_value.BestResponse(
        game, dist, value.TabularValueFunction(game))
    br_val = br_value(game.new_initial_state())
    self.assertAlmostEqual(br_val, 30.029387484327486)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2018 The Dopamine Authors and Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
# This file is a fork of the original Dopamine code incorporating changes for
# the multiplayer setting and the Hanabi Learning Environment.
#
"""Run methods for training a DQN agent on Atari.

Methods in this module are usually referenced by |train.py|.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time

from third_party.dopamine import checkpointer
from third_party.dopamine import iteration_statistics
import dqn_agent
import gin.tf
from hanabi_learning_environment import rl_env
import numpy as np
import rainbow_agent
import tensorflow as tf

LENIENT_SCORE = False


class ObservationStacker(object):
  """Class for stacking agent observations."""

  def __init__(self, history_size, observation_size, num_players):
    """Initializer for observation stacker.

    Args:
      history_size: int, number of time steps to stack.
      observation_size: int, size of observation vector on one time step.
      num_players: int, number of players.
    """
    self._history_size = history_size
    self._observation_size = observation_size
    self._num_players = num_players
    self._obs_stacks = list()
    for _ in range(0, self._num_players):
      self._obs_stacks.append(np.zeros(self._observation_size *
                                       self._history_size))

  def add_observation(self, observation, current_player):
    """Adds observation for the current player.

    Args:
      observation: observation vector for current player.
      current_player: int, current player id.
    """
    self._obs_stacks[current_player] = np.roll(self._obs_stacks[current_player],
                                               -self._observation_size)
    self._obs_stacks[current_player][(self._history_size - 1) *
                                     self._observation_size:] = observation

  def get_observation_stack(self, current_player):
    """Returns the stacked observation for current player.

    Args:
      current_player: int, current player id.
    """

    return self._obs_stacks[current_player]

  def reset_stack(self):
    """Resets the observation stacks to all zero."""

    for i in range(0, self._num_players):
      self._obs_stacks[i].fill(0.0)

  @property
  def history_size(self):
    """Returns number of steps to stack."""
    return self._history_size

  def observation_size(self):
    """Returns the size of the observation vector after history stacking."""
    return self._observation_size * self._history_size


def load_gin_configs(gin_files, gin_bindings):
  """Loads gin configuration files.

  Args:
    gin_files: A list of paths to the gin configuration files for this
      experiment.
    gin_bindings: List of gin parameter bindings to override the values in the
      config files.
  """
  gin.parse_config_files_and_bindings(gin_files,
                                      bindings=gin_bindings,
                                      skip_unknown=False)


@gin.configurable
def create_environment(game_type='Hanabi-Full', num_players=2):
  """Creates the Hanabi environment.

  Args:
    game_type: Type of game to play. Currently the following are supported:
      Hanabi-Full: Regular game.
      Hanabi-Small: The small version of Hanabi, with 2 cards and 2 colours.
    num_players: Int, number of players to play this game.

  Returns:
    A Hanabi environment.
  """
  return rl_env.make(
      environment_name=game_type, num_players=num_players, pyhanabi_path=None)


@gin.configurable
def create_obs_stacker(environment, history_size=4):
  """Creates an observation stacker.

  Args:
    environment: environment object.
    history_size: int, number of steps to stack.

  Returns:
    An observation stacker object.
  """

  return ObservationStacker(history_size,
                            environment.vectorized_observation_shape()[0],
                            environment.players)


@gin.configurable
def create_agent(environment, obs_stacker, agent_type='DQN'):
  """Creates the Hanabi agent.

  Args:
    environment: The environment.
    obs_stacker: Observation stacker object.
    agent_type: str, type of agent to construct.

  Returns:
    An agent for playing Hanabi.

  Raises:
    ValueError: if an unknown agent type is requested.
  """
  if agent_type == 'DQN':
    return dqn_agent.DQNAgent(observation_size=obs_stacker.observation_size(),
                              num_actions=environment.num_moves(),
                              num_players=environment.players)
  elif agent_type == 'Rainbow':
    return rainbow_agent.RainbowAgent(
        observation_size=obs_stacker.observation_size(),
        num_actions=environment.num_moves(),
        num_players=environment.players)
  else:
    raise ValueError('Expected valid agent_type, got {}'.format(agent_type))


def initialize_checkpointing(agent, experiment_logger, checkpoint_dir,
                             checkpoint_file_prefix='ckpt'):
  """Reloads the latest checkpoint if it exists.

  The following steps will be taken:
   - This method will first create a Checkpointer object, which will be used in
     the method and then returned to the caller for later use.
   - It will then call checkpointer.get_latest_checkpoint_number to determine
     whether there is a valid checkpoint in checkpoint_dir, and what is the
     largest file number.
   - If a valid checkpoint file is found, it will load the bundled data from
     this file and will pass it to the agent for it to reload its data.
   - If the agent is able to successfully unbundle, this method will verify that
     the unbundled data contains the keys, 'logs' and 'current_iteration'. It
     will then load the Logger's data from the bundle, and will return the
     iteration number keyed by 'current_iteration' as one of the return values
     (along with the Checkpointer object).

  Args:
    agent: The agent that will unbundle the checkpoint from checkpoint_dir.
    experiment_logger: The Logger object that will be loaded from the
      checkpoint.
    checkpoint_dir: str, the directory containing the checkpoints.
    checkpoint_file_prefix: str, the checkpoint file prefix.

  Returns:
    start_iteration: int, The iteration number to start the experiment from.
    experiment_checkpointer: The experiment checkpointer.
  """
  experiment_checkpointer = checkpointer.Checkpointer(
      checkpoint_dir, checkpoint_file_prefix)

  start_iteration = 0

  # Check if checkpoint exists. Note that the existence of checkpoint 0 means
  # that we have finished iteration 0 (so we will start from iteration 1).
  latest_checkpoint_version = checkpointer.get_latest_checkpoint_number(
      checkpoint_dir)
  if latest_checkpoint_version >= 0:
    dqn_dictionary = experiment_checkpointer.load_checkpoint(
        latest_checkpoint_version)
    if agent.unbundle(
        checkpoint_dir, latest_checkpoint_version, dqn_dictionary):
      assert 'logs' in dqn_dictionary
      assert 'current_iteration' in dqn_dictionary
      experiment_logger.data = dqn_dictionary['logs']
      start_iteration = dqn_dictionary['current_iteration'] + 1
      tf.logging.info('Reloaded checkpoint and will start from iteration %d',
                      start_iteration)

  return start_iteration, experiment_checkpointer


def format_legal_moves(legal_moves, action_dim):
  """Returns formatted legal moves.

  This function takes a list of actions and converts it into a fixed size vector
  of size action_dim. If an action is legal, its position is set to 0 and -Inf
  otherwise.
  Ex: legal_moves = [0, 1, 3], action_dim = 5
      returns [0, 0, -Inf, 0, -Inf]

  Args:
    legal_moves: list of legal actions.
    action_dim: int, number of actions.

  Returns:
    a vector of size action_dim.
  """
  new_legal_moves = np.full(action_dim, -float('inf'))
  if legal_moves:
    new_legal_moves[legal_moves] = 0
  return new_legal_moves


def parse_observations(observations, num_actions, obs_stacker):
  """Deconstructs the rich observation data into relevant components.

  Args:
    observations: dict, containing full observations.
    num_actions: int, The number of available actions.
    obs_stacker: Observation stacker object.

  Returns:
    current_player: int, Whose turn it is.
    legal_moves: `np.array` of floats, of length num_actions, whose elements
      are -inf for indices corresponding to illegal moves and 0, for those
      corresponding to legal moves.
    observation_vector: Vectorized observation for the current player.
  """
  current_player = observations['current_player']
  current_player_observation = (
      observations['player_observations'][current_player])

  legal_moves = current_player_observation['legal_moves_as_int']
  legal_moves = format_legal_moves(legal_moves, num_actions)

  observation_vector = current_player_observation['vectorized']
  obs_stacker.add_observation(observation_vector, current_player)
  observation_vector = obs_stacker.get_observation_stack(current_player)

  return current_player, legal_moves, observation_vector


def run_one_episode(agent, environment, obs_stacker):
  """Runs the agent on a single game of Hanabi in self-play mode.

  Args:
    agent: Agent playing Hanabi.
    environment: The Hanabi environment.
    obs_stacker: Observation stacker object.

  Returns:
    step_number: int, number of actions in this episode.
    total_reward: float, undiscounted return for this episode.
  """
  obs_stacker.reset_stack()
  observations = environment.reset()
  current_player, legal_moves, observation_vector = (
      parse_observations(observations, environment.num_moves(), obs_stacker))
  action = agent.begin_episode(current_player, legal_moves, observation_vector)

  is_done = False
  total_reward = 0
  step_number = 0

  has_played = {current_player}

  # Keep track of per-player reward.
  reward_since_last_action = np.zeros(environment.players)

  while not is_done:
    observations, reward, is_done, _ = environment.step(action.item())

    modified_reward = max(reward, 0) if LENIENT_SCORE else reward
    total_reward += modified_reward

    reward_since_last_action += modified_reward

    step_number += 1
    if is_done:
      break
    current_player, legal_moves, observation_vector = (
        parse_observations(observations, environment.num_moves(), obs_stacker))
    if current_player in has_played:
      action = agent.step(reward_since_last_action[current_player],
                          current_player, legal_moves, observation_vector)
    else:
      # Each player begins the episode on their first turn (which may not be
      # the first move of the game).
      action = agent.begin_episode(current_player, legal_moves,
                                   observation_vector)
      has_played.add(current_player)

    # Reset this player's reward accumulator.
    reward_since_last_action[current_player] = 0

  agent.end_episode(reward_since_last_action)

  tf.logging.info('EPISODE: %d %g', step_number, total_reward)
  return step_number, total_reward


def run_one_phase(agent, environment, obs_stacker, min_steps, statistics,
                  run_mode_str):
  """Runs the agent/environment loop until a desired number of steps.

  Args:
    agent: Agent playing hanabi.
    environment: environment object.
    obs_stacker: Observation stacker object.
    min_steps: int, minimum number of steps to generate in this phase.
    statistics: `IterationStatistics` object which records the experimental
      results.
    run_mode_str: str, describes the run mode for this agent.

  Returns:
    The number of steps taken in this phase, the sum of returns, and the
      number of episodes performed.
  """
  step_count = 0
  num_episodes = 0
  sum_returns = 0.

  while step_count < min_steps:
    episode_length, episode_return = run_one_episode(agent, environment,
                                                     obs_stacker)
    statistics.append({
        '{}_episode_lengths'.format(run_mode_str): episode_length,
        '{}_episode_returns'.format(run_mode_str): episode_return
    })

    step_count += episode_length
    sum_returns += episode_return
    num_episodes += 1

  return step_count, sum_returns, num_episodes


@gin.configurable
def run_one_iteration(agent, environment, obs_stacker,
                      iteration, training_steps,
                      evaluate_every_n=100,
                      num_evaluation_games=100):
  """Runs one iteration of agent/environment interaction.

  An iteration involves running several episodes until a certain number of
  steps are obtained.

  Args:
    agent: Agent playing hanabi.
    environment: The Hanabi environment.
    obs_stacker: Observation stacker object.
    iteration: int, current iteration number, used as a global_step.
    training_steps: int, the number of training steps to perform.
    evaluate_every_n: int, frequency of evaluation.
    num_evaluation_games: int, number of games per evaluation.

  Returns:
    A dict containing summary statistics for this iteration.
  """
  start_time = time.time()

  statistics = iteration_statistics.IterationStatistics()

  # First perform the training phase, during which the agent learns.
  agent.eval_mode = False
  number_steps, sum_returns, num_episodes = (
      run_one_phase(agent, environment, obs_stacker, training_steps, statistics,
                    'train'))
  time_delta = time.time() - start_time
  tf.logging.info('Average training steps per second: %.2f',
                  number_steps / time_delta)

  average_return = sum_returns / num_episodes
  tf.logging.info('Average per episode return: %.2f', average_return)
  statistics.append({'average_return': average_return})

  # Also run an evaluation phase if desired.
  if evaluate_every_n is not None and iteration % evaluate_every_n == 0:
    episode_data = []
    agent.eval_mode = True
    # Collect episode data for all games.
    for _ in range(num_evaluation_games):
      episode_data.append(run_one_episode(agent, environment, obs_stacker))

    eval_episode_length, eval_episode_return = map(np.mean, zip(*episode_data))

    statistics.append({
        'eval_episode_lengths': eval_episode_length,
        'eval_episode_returns': eval_episode_return
    })
    tf.logging.info('Average eval. episode length: %.2f  Return: %.2f',
                    eval_episode_length, eval_episode_return)
  else:
    statistics.append({
        'eval_episode_lengths': -1,
        'eval_episode_returns': -1
    })

  return statistics.data_lists


def log_experiment(experiment_logger, iteration, statistics,
                   logging_file_prefix='log', log_every_n=1):
  """Records the results of the current iteration.

  Args:
    experiment_logger: A `Logger` object.
    iteration: int, iteration number.
    statistics: Object containing statistics to log.
    logging_file_prefix: str, prefix to use for the log files.
    log_every_n: int, specifies logging frequency.
  """
  if iteration % log_every_n == 0:
    experiment_logger['iter{:d}'.format(iteration)] = statistics
    experiment_logger.log_to_file(logging_file_prefix, iteration)


def checkpoint_experiment(experiment_checkpointer, agent, experiment_logger,
                          iteration, checkpoint_dir, checkpoint_every_n):
  """Checkpoint experiment data.

  Args:
    experiment_checkpointer: A `Checkpointer` object.
    agent: An RL agent.
    experiment_logger: a Logger object, to include its data in the checkpoint.
    iteration: int, iteration number for checkpointing.
    checkpoint_dir: str, the directory where to save checkpoints.
    checkpoint_every_n: int, the frequency for writing checkpoints.
  """
  if iteration % checkpoint_every_n == 0:
    agent_dictionary = agent.bundle_and_checkpoint(checkpoint_dir, iteration)
    if agent_dictionary:
      agent_dictionary['current_iteration'] = iteration
      agent_dictionary['logs'] = experiment_logger.data
      experiment_checkpointer.save_checkpoint(iteration, agent_dictionary)


@gin.configurable
def run_experiment(agent,
                   environment,
                   start_iteration,
                   obs_stacker,
                   experiment_logger,
                   experiment_checkpointer,
                   checkpoint_dir,
                   num_iterations=200,
                   training_steps=5000,
                   logging_file_prefix='log',
                   log_every_n=1,
                   checkpoint_every_n=1):
  """Runs a full experiment, spread over multiple iterations."""
  tf.logging.info('Beginning training...')
  if num_iterations <= start_iteration:
    tf.logging.warning('num_iterations (%d) < start_iteration(%d)',
                       num_iterations, start_iteration)
    return

  for iteration in range(start_iteration, num_iterations):
    start_time = time.time()
    statistics = run_one_iteration(agent, environment, obs_stacker, iteration,
                                   training_steps)
    tf.logging.info('Iteration %d took %d seconds', iteration,
                    time.time() - start_time)
    start_time = time.time()
    log_experiment(experiment_logger, iteration, statistics,
                   logging_file_prefix, log_every_n)
    tf.logging.info('Logging iteration %d took %d seconds', iteration,
                    time.time() - start_time)
    start_time = time.time()
    checkpoint_experiment(experiment_checkpointer, agent, experiment_logger,
                          iteration, checkpoint_dir, checkpoint_every_n)
    tf.logging.info('Checkpointing iteration %d took %d seconds', iteration,
                    time.time() - start_time)
